from __future__ import annotations

import os
import unittest

import numpy as np
from dgl.data.utils import split_dataset
from pymatgen.core import Molecule
from pymatgen.util.testing import PymatgenTest

from matgl.ext.pymatgen import Molecule2Graph, Structure2Graph, get_element_list
from matgl.graph.data import (
    M3GNetDataset,
    MEGNetDataset,
    MGLDataLoader,
    collate_fn,
)

module_dir = os.path.dirname(os.path.abspath(__file__))


class DatasetTest(PymatgenTest):
    def test_megnet_dataset(self):
        s1 = self.get_structure("LiFePO4")
        s2 = self.get_structure("BaNiO3")
        structures = [s1, s2]
        label = [-1.0, 2.0]
        element_types = get_element_list([s1, s2])
        cry_graph = Structure2Graph(element_types=element_types, cutoff=4.0)
        dataset = MEGNetDataset(structures=structures, converter=cry_graph, labels=label, label_name="label")
        g1, label1, state1 = dataset[0]
        g2, label2, state2 = dataset[1]
        assert label1 == label[0]
        assert g1.num_edges() == cry_graph.get_graph(s1)[0].num_edges()
        assert g1.num_nodes() == cry_graph.get_graph(s1)[0].num_nodes()
        assert g2.num_edges() == cry_graph.get_graph(s2)[0].num_edges()
        assert g2.num_nodes() == cry_graph.get_graph(s2)[0].num_nodes()

    def test_megnet_dataset_for_mol(self):
        coords = [
            [0.000000, 0.000000, 0.000000],
            [0.000000, 0.000000, 1.089000],
            [1.026719, 0.000000, -0.363000],
            [-0.513360, -0.889165, -0.363000],
            [-0.513360, 0.889165, -0.363000],
        ]
        methane = Molecule(["C", "H", "H", "H", "H"], coords)
        element_types = get_element_list([methane])
        mol_graph = Molecule2Graph(element_types=element_types, cutoff=1.5)
        label = [1.0, 2.0]
        structures = [methane, methane]
        dataset = MEGNetDataset(
            structures=structures, converter=mol_graph, labels=label, label_name="label", name="MolDataset"
        )
        g1, label1, state1 = dataset[0]
        g2, label2, state2 = dataset[1]
        assert label1 == label[0]
        assert g1.num_edges() == mol_graph.get_graph(methane)[0].num_edges()
        assert g1.num_nodes() == mol_graph.get_graph(methane)[0].num_nodes()
        assert g2.num_edges() == mol_graph.get_graph(methane)[0].num_edges()
        assert g2.num_nodes() == mol_graph.get_graph(methane)[0].num_nodes()

    def test_m3gnet_dataset(self):
        s1 = self.get_structure("LiFePO4")
        s2 = self.get_structure("BaNiO3")
        structures = [s1, s2]
        energies = [-1.0, 2.0]
        forces = [np.zeros((28, 3)).tolist(), np.zeros((10, 3)).tolist()]
        stresses = [np.zeros((3, 3)).tolist(), np.zeros((3, 3)).tolist()]
        element_types = get_element_list([s1, s2])
        cry_graph = Structure2Graph(element_types=element_types, cutoff=4.0)
        dataset = M3GNetDataset(
            structures=structures,
            converter=cry_graph,
            threebody_cutoff=4.0,
            energies=energies,
            forces=forces,
            stresses=stresses,
        )
        g1, l_g1, state1, energies_g1, forces_g1, stresses_g1 = dataset[0]
        g2, l_g2, state2, energies_g2, forces_g2, stresses_g2 = dataset[1]
        assert energies_g1 == energies[0]
        assert g1.num_edges() == cry_graph.get_graph(s1)[0].num_edges()
        assert g1.num_nodes() == cry_graph.get_graph(s1)[0].num_nodes()
        assert g2.num_edges() == cry_graph.get_graph(s2)[0].num_edges()
        assert g2.num_nodes() == cry_graph.get_graph(s2)[0].num_nodes()
        assert np.shape(forces_g1)[0], 28
        assert np.shape(forces_g2)[0], 10

    def test_megnet_dataloader(self):
        s1 = self.get_structure("LiFePO4")
        s2 = self.get_structure("BaNiO3")
        structures = [s1, s1, s1, s1, s1, s1, s1, s1, s1, s1, s2, s2, s2, s2, s2, s2, s2, s2, s2, s2]
        label = np.zeros(20)
        element_types = get_element_list([s1, s2])
        cry_graph = Structure2Graph(element_types=element_types, cutoff=4.0)
        dataset = MEGNetDataset(structures=structures, converter=cry_graph, labels=label, label_name="label")
        train_data, val_data, test_data = split_dataset(
            dataset,
            frac_list=[0.8, 0.1, 0.1],
            shuffle=True,
            random_state=42,
        )
        train_loader, val_loader, test_loader = MGLDataLoader(
            train_data=train_data,
            val_data=val_data,
            test_data=test_data,
            collate_fn=collate_fn,
            batch_size=2,
            num_workers=1,
        )
        assert len(train_loader) == 8
        assert len(val_loader) == 1
        assert len(test_loader) == 1

    def test_megnet_dataloader_for_mol(self):
        coords = [
            [0.000000, 0.000000, 0.000000],
            [0.000000, 0.000000, 1.089000],
            [1.026719, 0.000000, -0.363000],
            [-0.513360, -0.889165, -0.363000],
            [-0.513360, 0.889165, -0.363000],
        ]
        m1 = Molecule(["C", "H", "H", "H", "H"], coords)
        structures = [m1, m1, m1, m1, m1, m1, m1, m1, m1, m1]
        label = np.zeros(10)
        element_types = get_element_list([m1])
        mol_graph = Molecule2Graph(element_types=element_types, cutoff=1.5)
        dataset = MEGNetDataset(structures=structures, converter=mol_graph, labels=label, label_name="label")
        train_data, val_data, test_data = split_dataset(
            dataset,
            frac_list=[0.6, 0.2, 0.2],
            shuffle=True,
            random_state=42,
        )
        train_loader, val_loader, test_loader = MGLDataLoader(
            train_data=train_data,
            val_data=val_data,
            test_data=test_data,
            collate_fn=collate_fn,
            batch_size=2,
            num_workers=1,
        )
        assert len(train_loader) == 3
        assert len(val_loader) == 1
        assert len(test_loader) == 1

    def test_m3gnet_dataloader(self):
        s1 = self.get_structure("LiFePO4")
        s2 = self.get_structure("BaNiO3")
        structures = [s1, s2, s1, s2, s1, s2, s1, s2, s1, s2, s1, s2, s1, s2, s1, s2, s1, s2, s1, s2]
        energies = np.zeros(20)
        f1 = np.zeros((28, 3)).tolist()
        f2 = np.zeros((10, 3)).tolist()
        s = np.zeros((3, 3)).tolist()
        forces = [f1, f2, f1, f2, f1, f2, f1, f2, f1, f2, f1, f2, f1, f2, f1, f2, f1, f2, f1, f2]
        stresses = [s, s, s, s, s, s, s, s, s, s, s, s, s, s, s, s, s, s, s, s]
        element_types = get_element_list([s1, s2])
        cry_graph = Structure2Graph(element_types=element_types, cutoff=4.0)
        dataset = M3GNetDataset(
            structures=structures,
            converter=cry_graph,
            threebody_cutoff=4.0,
            energies=energies,
            forces=forces,
            stresses=stresses,
        )
        train_data, val_data, test_data = split_dataset(
            dataset,
            frac_list=[0.8, 0.1, 0.1],
            shuffle=True,
            random_state=42,
        )
        train_loader, val_loader, test_loader = MGLDataLoader(
            train_data=train_data,
            val_data=val_data,
            test_data=test_data,
            collate_fn=collate_fn,
            batch_size=2,
            num_workers=1,
        )
        assert len(train_loader) == 8
        assert len(val_loader) == 1
        assert len(test_loader) == 1

    @classmethod
    def tearDownClass(cls):
        for fn in ("dgl_graph.bin", "dgl_line_graph.bin", "state_attr.pt", "labels.json"):
            try:
                os.remove(fn)
            except FileNotFoundError:
                pass


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import unittest

from pymatgen.core.structure import Lattice, Molecule, Structure

from matgl.ext.pymatgen import Molecule2Graph, Structure2Graph, get_element_list
from matgl.graph.compute import compute_pair_vector_and_distance
from matgl.layers import BondExpansion


class TestBondExpansion(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.s = Structure(Lattice.cubic(4.0), ["Mo", "S"], [[0.01, 0.0, 0.0], [0.5, 0.5, 0.5]])
        mol = Molecule(["C", "O"], [[0, 0, 0], [1.1, 0, 0]])

        element_types = get_element_list([cls.s])
        p2g = Structure2Graph(element_types=element_types, cutoff=4.0)  # type: ignore
        graph, state = p2g.get_graph(cls.s)
        cls.g1 = graph
        cls.state1 = state

        element_types = get_element_list([mol])
        p2g = Molecule2Graph(element_types=element_types, cutoff=4.0)  # type: ignore
        graph, state = p2g.get_graph(mol)
        cls.g2 = graph
        cls.state2 = state

    def test_gaussian(self):
        bond_expansion = BondExpansion(rbf_type="Gaussian", num_centers=10, initial=0.0, final=4.0, width=0.5)
        bond_vec, bond_dist = compute_pair_vector_and_distance(self.g1)
        bond_basis = bond_expansion(bond_dist)
        assert bond_basis.shape == (28, 10)

        bond_vec, bond_dist = compute_pair_vector_and_distance(self.g2)
        bond_basis = bond_expansion(bond_dist)
        assert bond_basis.shape == (2, 10)

    def test_spherical_bessel_with_smooth(self):
        bond_expansion = BondExpansion(rbf_type="SphericalBessel", max_n=3, max_l=3, cutoff=4.0, smooth=True)
        bond_vec, bond_dist = compute_pair_vector_and_distance(self.g1)
        bond_basis = bond_expansion(bond_dist)
        assert bond_basis.shape == (28, 3)

        bond_vec, bond_dist = compute_pair_vector_and_distance(self.g2)
        bond_basis = bond_expansion(bond_dist)
        assert bond_basis.shape == (2, 3)

    def test_spherical_bessel(self):
        bond_expansion = BondExpansion(rbf_type="SphericalBessel", max_n=3, max_l=3, cutoff=4.0, smooth=False)
        bond_vec, bond_dist = compute_pair_vector_and_distance(self.g1)
        bond_basis = bond_expansion(bond_dist)
        assert bond_basis.shape == (28, 9)

        bond_vec, bond_dist = compute_pair_vector_and_distance(self.g2)
        bond_basis = bond_expansion(bond_dist)
        assert bond_basis.shape == (2, 9)


if __name__ == "__main__":
    unittest.main()



"""
Generate bond features based on spherical bessel functions or gaussian expansion
"""
from __future__ import annotations

import torch
from torch import nn

from matgl.utils.maths import GaussianExpansion, SphericalBesselFunction


class BondExpansion(nn.Module):
    """
    Expand pair distances into a set of spherical bessel or gaussian functions.
    """

    def __init__(
        self,
        max_l: int = 3,
        max_n: int = 3,
        cutoff: float = 5.0,
        rbf_type: str = "SphericalBessel",
        smooth: bool = False,
        initial: float = 0.0,
        final: float = 5.0,
        num_centers: int = 100,
        width: float = 0.5,
    ) -> None:
        """
        Args:
            max_l (int): order of angular part
            max_n (int): order of radial part
            cutoff (float): cutoff radius
            rbf_type (str): type of radial basis function .i.e. either "SphericalBessel" or 'Gaussian'
            smooth (bool): whether apply the smooth version of spherical bessel functions or not
            initial (float): initial point for gaussian expansion
            final (float): final point for gaussian expansion
            num_centers (int): Number of centers for gaussian expansion.
            width (float): width of gaussian function
        """
        super().__init__()

        self.max_n = max_n
        self.cutoff = cutoff
        self.max_l = max_l
        self.smooth = smooth
        self.num_centers = num_centers
        self.width = width
        self.initial = initial
        self.final = final
        self.rbf_type = rbf_type

        if rbf_type == "SphericalBessel":
            self.rbf = SphericalBesselFunction(max_l, max_n, cutoff, smooth)  # type: ignore
        elif rbf_type == "Gaussian":
            self.rbf = GaussianExpansion(initial, final, num_centers, width)  # type: ignore
        else:
            raise Exception("undefined rbf_type, please use SphericalBessel or Gaussian instead.")

    def forward(self, bond_dist: torch.tensor):
        """
        Forward

        Args:
        bond_dist: Bond distance

        Return:
        bond_basis: Radial basis functions
        """
        bond_basis = self.rbf(bond_dist)
        return bond_basis



"""
Implementations of pseudomodels that wraps other models.
"""
from __future__ import annotations

from torch import nn

from matgl.data.transformer import Transformer
from matgl.utils.io import IOMixIn


class TransformedTargetModel(nn.Module, IOMixIn):
    """
    A model where the target is first transformed prior to training and the reverse transformation is performed for
    predictions. This is modelled after scikit-learn's TransformedTargetRegressor
    """

    def __init__(self, model: nn.Module, target_transformer: Transformer):
        """
        Args:
            model: Input model
            target_transformer: Transformer for target.
        """
        super().__init__()
        self.save_args(locals())
        self.model = model
        self.transformer = target_transformer

    def forward(self, *args, **kwargs):
        r"""
        Args:
            *args: Passthrough to parent model.forward method.
            **kwargs: Passthrough to parent model.forward method.

        Returns:
            Inverse transformed output.
        """
        output = self.model.forward(*args, **kwargs)
        return self.transformer.inverse_transform(output)

    def __repr__(self):
        return f"TransformedTargetModel:\nModel: {self.model.__repr()}\nTransformer: {self.transformer.__repr__()}"

    def predict_structure(self, *args, **kwargs):
        """
        Pass through to parent model.predict_structure with inverse transform.

        Args:
            *args: Pass-through to self.model.predict_structure.
            **kwargs: Pass-through to self.model.predict_structure.

        Returns:
            Transformed answer.
        """
        return self.transformer.inverse_transform(self.model.predict_structure(*args, **kwargs))



from __future__ import annotations

import unittest
from collections import namedtuple

import dgl
import numpy as np
import torch
from pymatgen.core.structure import Lattice, Structure
from torch import nn

from matgl.ext.pymatgen import Structure2Graph, get_element_list
from matgl.graph.compute import (
    compute_pair_vector_and_distance,
    compute_theta_and_phi,
    create_line_graph,
)
from matgl.layers import BondExpansion, EmbeddingBlock
from matgl.layers._graph_convolution import (
    MLP,
    M3GNetBlock,
    M3GNetGraphConv,
    MEGNetBlock,
    MEGNetGraphConv,
)
from matgl.layers._three_body import SphericalBesselWithHarmonics
from matgl.utils.cutoff import polynomial_cutoff

Graph = namedtuple("Graph", "graph, state_attr")


def build_graph(N, E, NDIM=5, EDIM=3, GDIM=10):
    graph = dgl.rand_graph(N, E)
    graph.ndata["node_feat"] = torch.rand(N, NDIM)
    graph.edata["edge_feat"] = torch.rand(E, EDIM)
    state_attr = torch.rand(1, GDIM)
    return Graph(graph, state_attr)


def get_graphs(num_graphs, NDIM=5, EDIM=3, GDIM=10):
    Ns = torch.randint(10, 30, (num_graphs,)).tolist()
    Es = torch.randint(35, 100, (num_graphs,)).tolist()
    graphs = [build_graph(*gspec, NDIM, EDIM, GDIM) for gspec in zip(Ns, Es)]
    return graphs


def batch(state_attrs_lists):
    graphs, attrs = list(zip(*state_attrs_lists))
    batched_graph = dgl.batch(graphs)
    batched_attrs = torch.vstack(attrs)
    return batched_graph, batched_attrs


def test_megnet_layer():
    graphs = get_graphs(5)
    batched_graph, attrs = batch(graphs)

    NDIM, EDIM, GDIM = 5, 3, 10
    edge_func = MLP(dims=[2 * NDIM + EDIM + GDIM, EDIM])
    node_func = MLP(dims=[EDIM + NDIM + GDIM, NDIM])
    state_func = MLP(dims=[EDIM + NDIM + GDIM, GDIM])
    layer = MEGNetGraphConv(edge_func, node_func, state_func)

    # one pass
    edge_feat = batched_graph.edata.pop("edge_feat")
    node_feat = batched_graph.ndata.pop("node_feat")
    out = layer(batched_graph, edge_feat, node_feat, attrs)
    return out


def test_megnet_block():
    DIM = 5
    N1, N2 = 64, 32
    block = MEGNetBlock(
        dims=[5, 10, 13],
        conv_hiddens=[N1, N1, N2],
        act=nn.SiLU(),
        skip=False,
    )
    graphs = get_graphs(5, NDIM=DIM, EDIM=DIM, GDIM=DIM)
    batched_graph, attrs = batch(graphs)

    # one pass
    edge_feat = batched_graph.edata.pop("edge_feat")
    node_feat = batched_graph.ndata.pop("node_feat")
    out = block(batched_graph, edge_feat, node_feat, attrs)
    return out


class TestGraphConv(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.s = Structure(Lattice.cubic(4.0), ["Mo", "S"], [[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]])
        cls.s.states = np.array([[0.1, 0.2, 0.3, 0.4, 0.5]])

        element_types = get_element_list([cls.s])
        p2g = Structure2Graph(element_types=element_types, cutoff=4.0)
        graph, state = p2g.get_graph(cls.s)
        cls.g1 = graph
        cls.state1 = state
        bond_vec, bond_dist = compute_pair_vector_and_distance(cls.g1)
        cls.g1.edata["bond_dist"] = bond_dist
        cls.g1.edata["bond_vec"] = bond_vec

    def test_m3gnet_graph_conv(self):
        bond_dist = self.g1.edata["bond_dist"]
        polynomial_cutoff(bond_dist, 4.0)
        bond_expansion = BondExpansion(max_l=3, max_n=3, cutoff=5.0, rbf_type="SphericalBessel", smooth=False)
        bond_basis = bond_expansion(bond_dist)
        self.g1.edata["rbf"] = bond_basis
        sb_and_sh = SphericalBesselWithHarmonics(max_n=3, max_l=3, cutoff=5.0, use_smooth=False, use_phi=False)
        l_g1 = create_line_graph(self.g1, threebody_cutoff=4.0)
        l_g1.apply_edges(compute_theta_and_phi)
        sb_and_sh(l_g1)
        max_n = 3
        max_l = 3
        num_node_feats = 16
        num_edge_feats = 24
        num_state_feats = 32
        node_attr = self.g1.ndata["attr"]
        state_attr = torch.tensor([0.0, 0.0])
        embedding = EmbeddingBlock(
            degree_rbf=9,
            dim_node_embedding=num_node_feats,
            dim_edge_embedding=num_edge_feats,
            dim_state_feats=num_state_feats,
            include_state=True,
            activation=nn.SiLU(),
        )
        node_feat, edge_feat, state_feat = embedding(node_attr, bond_basis, state_attr)
        node_dim = node_feat.shape[-1]
        edge_dim = edge_feat.shape[-1]
        state_dim = state_feat.shape[-1]

        conv_hiddens = [32, 32]
        edge_in = 2 * node_dim + edge_dim + state_dim
        node_in = 2 * node_dim + edge_dim + state_dim
        state_in = node_dim + state_dim
        degree = max_n * max_l
        conv = M3GNetGraphConv.from_dims(
            degree=degree,
            include_states=True,
            edge_dims=[edge_in, *conv_hiddens, num_edge_feats],
            node_dims=[node_in, *conv_hiddens, num_node_feats],
            state_dims=[state_in, *conv_hiddens, num_state_feats],
            activation=nn.SiLU(),
        )
        edge_feat_new, node_feat_new, state_feat_new = conv(self.g1, edge_feat, node_feat, state_feat)
        assert [edge_feat_new.size(dim=0), edge_feat_new.size(dim=1)] == [28, 24]
        assert [node_feat_new.size(dim=0), node_feat_new.size(dim=1)] == [2, 16]
        assert [state_feat_new.size(dim=0), state_feat_new.size(dim=1)] == [1, 32]

    def test_m3gnet_block(self):
        polynomial_cutoff(self.g1.edata["bond_dist"], 4.0)
        bond_expansion = BondExpansion(max_l=3, max_n=3, cutoff=5.0, rbf_type="SphericalBessel", smooth=False)
        bond_basis = bond_expansion(self.g1.edata["bond_dist"])
        self.g1.edata["rbf"] = bond_basis
        sb_and_sh = SphericalBesselWithHarmonics(max_n=3, max_l=3, cutoff=5.0, use_smooth=False, use_phi=False)
        l_g1 = create_line_graph(self.g1, threebody_cutoff=4.0)
        l_g1.apply_edges(compute_theta_and_phi)
        sb_and_sh(l_g1)
        num_node_feats = 16
        num_edge_feats = 32
        num_state_feats = 64
        state_attr = torch.tensor([0.0, 0.0])
        embedding = EmbeddingBlock(
            degree_rbf=9,
            dim_node_embedding=num_node_feats,
            dim_edge_embedding=num_edge_feats,
            dim_state_feats=num_state_feats,
            include_state=True,
            activation=nn.SiLU(),
        )
        node_attr = self.g1.ndata["attr"]
        edge_attr = bond_basis
        node_feat, edge_feat, state_feat = embedding(node_attr, edge_attr, state_attr)
        self.g1.ndata["node_feat"] = node_feat
        self.g1.edata["edge_feat"] = edge_feat
        graph_conv = M3GNetBlock(
            degree=3 * 3,
            activation=nn.SiLU(),
            conv_hiddens=[32, 16],
            num_node_feats=num_node_feats,
            num_edge_feats=num_edge_feats,
            num_state_feats=num_state_feats,
            include_state=True,
        )
        edge_feat_new, node_feat_new, state_feat_new = graph_conv(self.g1, edge_feat, node_feat, state_feat)
        assert [edge_feat_new.size(dim=0), edge_feat_new.size(dim=1)] == [28, 32]
        assert [node_feat_new.size(dim=0), node_feat_new.size(dim=1)] == [2, 16]
        assert [state_feat_new.size(dim=0), state_feat_new.size(dim=1)] == [1, 64]

        # without state features
        state_feat = None
        graph_conv = M3GNetBlock(
            degree=3 * 3,
            num_node_feats=num_node_feats,
            num_edge_feats=num_edge_feats,
            num_state_feats=num_state_feats,
            conv_hiddens=[32, 16],
            activation=nn.SiLU(),
            include_state=False,
        )
        edge_feat_new, node_feat_new, state_feat_new = graph_conv(self.g1, edge_feat, node_feat, state_feat)
        assert [edge_feat_new.size(dim=0), edge_feat_new.size(dim=1)] == [28, 32]
        assert [node_feat_new.size(dim=0), node_feat_new.size(dim=1)] == [2, 16]


if __name__ == "__main__":
    test_megnet_layer()
    test_megnet_block()
    unittest.main()



from __future__ import annotations

import os
import unittest

import numpy as np
import torch
from pymatgen.core.structure import Lattice, Structure

from matgl.ext.pymatgen import Structure2Graph, get_element_list
from matgl.models._m3gnet import M3GNet


class TestM3GNet(unittest.TestCase):
    element_types = None
    g1 = None
    state1 = None

    @classmethod
    def setUpClass(cls) -> None:
        s = Structure(Lattice.cubic(4.0), ["Mo", "S"], [[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]])
        s.states = np.array([[0.1, 0.2, 0.3, 0.4, 0.5]])

        cls.element_types = get_element_list([s])
        p2g = Structure2Graph(element_types=cls.element_types, cutoff=5.0)
        graph, state = p2g.get_graph(s)
        cls.g1 = graph
        cls.state1 = state

    def test_model(self):
        model = M3GNet(element_types=self.element_types, is_intensive=False)
        output = model(g=self.g1)
        assert torch.numel(output) == 1
        model.save(".")
        M3GNet.load(".")
        os.remove("model.pt")
        os.remove("model.json")
        os.remove("state.pt")

    def test_model_intensive(self):
        model = M3GNet(element_types=self.element_types, is_intensive=True)
        output = model(g=self.g1)
        assert torch.numel(output) == 1

    def test_model_intensive_with_classification(self):
        model = M3GNet(
            element_types=self.element_types,
            is_intensive=True,
            task_type="classification",
        )
        output = model(g=self.g1)
        assert torch.numel(output) == 1

    def test_model_intensive_set2set_classification(self):
        model = M3GNet(
            element_types=self.element_types, is_intensive=True, task_type="classification", readout_type="set2set"
        )
        output = model(g=self.g1)
        assert torch.numel(output) == 1


if __name__ == "__main__":
    unittest.main()



"""
atomic energy offset. Used for predicting extensive properties.
"""
from __future__ import annotations

import dgl
import numpy as np
import torch
from pymatgen.core import Molecule, Structure
from torch import nn


class AtomRef(nn.Module):
    """
    Get total property offset for a system:
    """

    def __init__(
        self,
        property_offset: np.array,  # type: ignore
    ) -> None:
        """
        Args:
        -----------
        property_offset (np.array): a array of elemental property offset
        """
        super().__init__()
        self.property_offset = torch.tensor(property_offset)
        self.max_z = self.property_offset.size(dim=0)

    def get_feature_matrix(self, structs_or_graphs: list, element_list: tuple[str]) -> np.typing.NDArray:
        """
        Get the number of atoms for different elements in the structure

        Args:
        structs_or_graphs (list): a list of pymatgen Structure or dgl graph
        element_list: a dictionary containing element types in the training set

        Returns:
        features (np.array): a matrix (num_structures, num_elements)
        """
        n = len(structs_or_graphs)
        features = np.zeros(shape=(n, self.max_z))
        for i, s in enumerate(structs_or_graphs):
            if isinstance(s, (Structure, Molecule)):
                atomic_numbers = [element_list.index(site.specie.symbol) for site in s.sites]
            else:
                one_hot_vecs = s.ndata["attr"]
                atomic_numbers = ((one_hot_vecs == 1).nonzero(as_tuple=True)[0]).tolist()
            features[i] = np.bincount(atomic_numbers, minlength=self.max_z)
        return features

    def fit(self, structs_or_graphs: list, element_list: tuple[str], properties: np.typing.NDArray) -> None:
        """
        Fit the elemental reference values for the properties

        Args:
            structs_or_graphs: pymatgen Structures or dgl graphs
            element_list (tuple): a list of element types
            properties (np.ndarray): array of extensive properties
        """
        features = self.get_feature_matrix(structs_or_graphs, element_list)
        self.property_offset = np.linalg.pinv(features.T.dot(features)).dot(features.T.dot(properties))
        self.property_offset = torch.tensor(self.property_offset)

    def forward(self, g: dgl.DGLGraph, state_attr: torch.tensor | None = None):
        """
        Get the total property offset for a system

        Args:
        g: a batch of dgl graphs
        state_attr: state attributes

        Returns:
        offset_per_graph:
        """
        if self.property_offset.ndim > 1:
            offset_batched_with_state = []
            for i in range(0, self.property_offset.size(dim=0)):
                property_offset_batched = self.property_offset[i].repeat(g.num_nodes(), 1)
                offset = property_offset_batched * g.ndata["attr"]
                g.ndata["atomic_offset"] = torch.sum(offset, 1)
                offset_batched = dgl.readout_nodes(g, "atomic_offset")
                offset_batched_with_state.append(offset_batched)
            offset_batched_with_state = torch.stack(offset_batched_with_state)  # type: ignore
            return offset_batched_with_state[state_attr]  # type: ignore
        property_offset_batched = self.property_offset.repeat(g.num_nodes(), 1)
        offset = property_offset_batched * g.ndata["attr"]
        g.ndata["atomic_offset"] = torch.sum(offset, 1)
        offset_batched = dgl.readout_nodes(g, "atomic_offset")
        return offset_batched



"""
Computing various graph based operations
"""
from __future__ import annotations

import dgl
import numpy as np
import torch


def compute_3body(g: dgl.DGLGraph):
    """
    Calculate the three body indices from pair atom indices

    Args:
        g: DGL graph

    Returns:
        l_g: DGL graph containing three body information from graph
        triple_bond_indices (np.ndarray): bond indices that form three-body
        n_triple_ij (np.ndarray): number of three-body angles for each bond
        n_triple_i (np.ndarray): number of three-body angles each atom
        n_triple_s (np.ndarray): number of three-body angles for each structure
    """
    bond_atom_indices = g.edges()
    n_atoms = [g.num_nodes()]
    n_atoms_total = np.sum(g.num_nodes())
    first_col = bond_atom_indices[0].reshape(-1, 1)
    all_indices = torch.arange(n_atoms_total).reshape(1, -1)
    n_bond_per_atom = torch.count_nonzero(first_col == all_indices, dim=0)
    n_triple_i = n_bond_per_atom * (n_bond_per_atom - 1)
    n_triple = torch.sum(n_triple_i)
    n_triple_ij = (n_bond_per_atom - 1).repeat_interleave(n_bond_per_atom)
    triple_bond_indices = torch.empty((n_triple, 2), dtype=torch.int64)

    start = 0
    cs = 0
    for n in n_bond_per_atom:
        if n > 0:
            """
            triple_bond_indices is generated from all pair permutations of atom indices. The
            numpy version below does this with much greater efficiency. The equivalent slow
            code is:

            ```
            for j, k in itertools.permutations(range(n), 2):
                triple_bond_indices[index] = [start + j, start + k]
            ```
            """
            r = torch.arange(n)
            x, y = torch.meshgrid(r, r)
            c = torch.stack([y.ravel(), x.ravel()], dim=1)
            final = c[c[:, 0] != c[:, 1]]
            triple_bond_indices[start : start + (n * (n - 1)), :] = final + cs
            start += n * (n - 1)
            cs += n

    n_triple_s = []
    i = 0
    for n in n_atoms:
        j = i + n
        n_triple_s.append(torch.sum(n_triple_i[i:j]))
        i = j

    src_id, dst_id = (triple_bond_indices[:, 0], triple_bond_indices[:, 1])
    l_g = dgl.graph((src_id, dst_id))
    l_g.ndata["bond_dist"] = g.edata["bond_dist"]
    l_g.ndata["bond_vec"] = g.edata["bond_vec"]
    l_g.ndata["pbc_offset"] = g.edata["pbc_offset"]
    l_g.ndata["n_triple_ij"] = n_triple_ij
    n_triple_s = torch.tensor(n_triple_s, dtype=torch.int64)
    return l_g, triple_bond_indices, n_triple_ij, n_triple_i, n_triple_s


def compute_pair_vector_and_distance(g: dgl.DGLGraph):
    """
    Calculate bond vectors and distances using dgl graphs

    Args:
    g: DGL graph

    Returns:
    bond_vec (torch.tensor): bond distance between two atoms
    bond_dist (torch.tensor): vector from src node to dst node
    """
    bond_vec = torch.zeros(g.num_edges(), 3)
    bond_vec[:, :] = (
        g.ndata["pos"][g.edges()[1][:].long(), :]
        + torch.squeeze(torch.matmul(g.edata["pbc_offset"].unsqueeze(1), torch.squeeze(g.edata["lattice"])))
        - g.ndata["pos"][g.edges()[0][:].long(), :]
    )

    bond_dist = torch.norm(bond_vec, dim=1)

    return bond_vec, bond_dist


def compute_theta_and_phi(edges: dgl.udf.EdgeBatch):
    """
    Calculate bond angle Theta and Phi using dgl graphs

    Args:
    edges: DGL graph edges

    Returns:
    cos_theta: torch.tensor
    phi: torch.tensor
    triple_bond_lengths (torch.tensor):
    """
    vec1 = edges.src["bond_vec"]
    vec2 = edges.dst["bond_vec"]
    cosine_theta = torch.sum(vec1 * vec2, dim=1) / (torch.norm(vec1, dim=1) * torch.norm(vec2, dim=1))
    return {
        "cos_theta": cosine_theta,
        "phi": torch.zeros_like(cosine_theta),
        "triple_bond_lengths": edges.dst["bond_dist"],
    }


def create_line_graph(g_batched: dgl.DGLGraph, threebody_cutoff: float):
    """
    Calculate the three body indices from pair atom indices

    Args:
        g_batched: Batched DGL graph
        threebody_cutoff (float): cutoff for three-body interactions

    Returns:
        l_g: DGL graph containing three body information from graph
    """
    g_unbatched = dgl.unbatch(g_batched)
    l_g_unbatched = []
    for g in g_unbatched:
        bond_atom_indices = g.edges()
        n_bond = bond_atom_indices[0].size(dim=0)
        if n_bond > 0 and threebody_cutoff is not None:
            valid_three_body = g.edata["bond_dist"] <= threebody_cutoff
            src_id_with_three_body = bond_atom_indices[0][valid_three_body]
            dst_id_with_three_body = bond_atom_indices[1][valid_three_body]
            graph_with_three_body = dgl.graph((src_id_with_three_body, dst_id_with_three_body))
            graph_with_three_body.edata["bond_dist"] = g.edata["bond_dist"][valid_three_body]
            graph_with_three_body.edata["bond_vec"] = g.edata["bond_vec"][valid_three_body]
            graph_with_three_body.edata["pbc_offset"] = g.edata["pbc_offset"][valid_three_body]
        if graph_with_three_body.edata["bond_dist"].size(dim=0) > 0:
            l_g, triple_bond_indices, n_triple_ij, n_triple_i, n_triple_s = compute_3body(graph_with_three_body)
            l_g_unbatched.append(l_g)

    l_g_batched = dgl.batch(l_g_unbatched)
    return l_g_batched



"""
Three-Body interaction implementations.
"""

from __future__ import annotations

import dgl
import torch
from torch import nn

from matgl.utils.maths import (
    SphericalBesselFunction,
    SphericalHarmonicsFunction,
    combine_sbf_shf,
    get_segment_indices_from_n,
    scatter_sum,
)


class SphericalBesselWithHarmonics(nn.Module):
    """
    Expansion of basis using Spherical Bessel and Harmonics
    """

    def __init__(self, max_n: int, max_l: int, cutoff: float, use_smooth: bool, use_phi: bool):
        """
        :param max_n: Degree of radial basis functions
        :param max_l: Degree of angular basis functions
        :param cutoff: Cutoff sphere
        :param use_smooth: Whether using smooth version of SBFs or not
        :param use_phi: using phi as angular basis functions
        """
        super().__init__()

        assert max_n <= 64
        self.max_n = max_n
        self.max_l = max_l
        self.cutoff = cutoff
        self.use_phi = use_phi
        self.use_smooth = use_smooth

        # retrieve formulas
        self.shf = SphericalHarmonicsFunction(self.max_l, self.use_phi)
        if self.use_smooth:
            self.sbf = SphericalBesselFunction(self.max_l, self.max_n * self.max_l, self.cutoff, self.use_smooth)
        else:
            self.sbf = SphericalBesselFunction(self.max_l, self.max_n, self.cutoff, self.use_smooth)

    def forward(self, line_graph):
        sbf = self.sbf(line_graph.edata["triple_bond_lengths"])
        shf = self.shf(line_graph.edata["cos_theta"], line_graph.edata["phi"])
        return combine_sbf_shf(sbf, shf, max_n=self.max_n, max_l=self.max_l, use_phi=self.use_phi)


class ThreeBodyInteractions(nn.Module):
    """
    Include 3D interactions to the bond update
    """

    def __init__(self, update_network_atom: nn.Module, update_network_bond: nn.Module, **kwargs):
        r"""
        Args:
            update_network_atom: MLP for node features in Eq.2
            update_network_bond: Gated-MLP for edge features in Eq.3
            **kwargs: Kwargs pass-through to nn.Module.__init__().
        """
        super().__init__(**kwargs)
        self.update_network_atom = update_network_atom
        self.update_network_bond = update_network_bond

    def forward(
        self,
        graph: dgl.DGLGraph,
        line_graph: dgl.DGLGraph,
        three_basis: torch.tensor,
        three_cutoff: float,
        node_feat: torch.tensor,
        edge_feat: torch.tensor,
    ):
        """
        Args:
            graph: dgl graph
            line_graph: line graph.
            three_basis: three body basis expansion
            three_cutoff: cutoff radius
            node_feat: node features
            edge_feat: edge features
        """
        end_atom_index = torch.gather(graph.edges()[1], 0, line_graph.edges()[1].to(torch.int64))
        atoms = self.update_network_atom(node_feat)
        end_atom_index = torch.unsqueeze(end_atom_index, 1)
        atoms = torch.squeeze(atoms[end_atom_index])
        basis = three_basis * atoms
        three_cutoff = torch.unsqueeze(three_cutoff, dim=1)
        weights = torch.reshape(
            three_cutoff[torch.stack(list(line_graph.edges()), dim=1).to(torch.int64)], (-1, 2)  # type: ignore
        )
        weights = torch.prod(weights, axis=-1)
        basis = basis * weights[:, None]
        new_bonds = scatter_sum(
            basis.to(torch.float32),
            segment_ids=get_segment_indices_from_n(line_graph.ndata["n_triple_ij"]),
            num_segments=graph.num_edges(),
            dim=0,
        )
        edge_feat_updated = edge_feat + self.update_network_bond(new_bonds)
        return edge_feat_updated



from __future__ import annotations

import unittest

import dgl
import numpy as np
import torch

from matgl.utils.maths import (
    SPHERICAL_BESSEL_ROOTS,
    GaussianExpansion,
    SphericalBesselFunction,
    SphericalHarmonicsFunction,
    broadcast_states_to_atoms,
    broadcast_states_to_bonds,
    combine_sbf_shf,
    get_range_indices_from_n,
    get_segment_indices_from_n,
    repeat_with_n,
    spherical_bessel_roots,
    spherical_bessel_smooth,
    unsorted_segment_fraction,
)


class TestMath(unittest.TestCase):
    def test_spherical_bessel_roots(self):
        roots = spherical_bessel_roots(max_l=1, max_n=5)
        roots2 = SPHERICAL_BESSEL_ROOTS
        assert np.allclose(roots2[0, :5], roots.ravel())

    def test_spherical_bessel_harmonics_function(self):
        r = torch.empty(10).normal_()
        sbf = SphericalBesselFunction(max_l=3, cutoff=5.0, max_n=3, smooth=False)
        res = sbf(r)
        res2 = sbf.rbf_j0(r, cutoff=5.0, max_n=3)
        assert np.allclose(res[:, :3].numpy().ravel(), res2.numpy().ravel(), atol=1e-07)

        assert res.numpy().shape == (10, 9)

        sbf2 = SphericalBesselFunction(max_l=3, cutoff=5.0, max_n=3, smooth=True)

        res2 = sbf2(r)
        assert tuple(res2.shape) == (10, 3)

        shf = SphericalHarmonicsFunction(max_l=3, use_phi=True)
        res_shf = shf(costheta=torch.linspace(-1, 1, 10), phi=torch.linspace(0, 2 * np.pi, 10))

        assert res_shf.numpy().shape == (10, 9)
        combined = combine_sbf_shf(res, res_shf, max_n=3, max_l=3, use_phi=True)

        assert combined.shape == (10, 27)

        res_shf2 = SphericalHarmonicsFunction(max_l=3, use_phi=False)(
            costheta=torch.linspace(-1, 1, 10), phi=torch.linspace(0, 2 * np.pi, 10)
        )
        combined = combine_sbf_shf(res, res_shf2, max_n=3, max_l=3, use_phi=False)

        assert combined.shape == (10, 9)
        rdf = spherical_bessel_smooth(r, cutoff=5.0, max_n=3)
        assert rdf.numpy().shape == (10, 3)

    def test_torch_operations(self):
        ns = torch.tensor([2, 3])
        assert [0, 0, 1, 1, 1] == get_segment_indices_from_n(ns).tolist()
        ns = torch.tensor([2, 3])
        assert [0, 1, 0, 1, 2] == get_range_indices_from_n(ns).tolist()
        assert repeat_with_n(torch.tensor([[0, 0], [1, 1], [2, 2]]), torch.tensor([1, 2, 3])).tolist() == [
            [0, 0],
            [1, 1],
            [1, 1],
            [2, 2],
            [2, 2],
            [2, 2],
        ]

    def test_segments(self):
        x = torch.tensor([1.0, 1.0, 2.0, 3.0])
        res = unsorted_segment_fraction(x, torch.tensor([0, 0, 1, 1]), 2)
        assert np.allclose(res.tolist(), [0.5, 0.5, 0.4, 0.6])

    #        res = unsorted_segment_softmax(x, torch.tensor([0, 0, 1, 1]), 2)
    #        np.testing.assert_array_almost_equal(res, [0.5, 0.5, 0.26894143, 0.7310586])

    def test_broadcast(self):
        src_ids = torch.tensor([2, 3, 4])
        dst_ids = torch.tensor([1, 2, 3])
        g = dgl.graph((src_ids, dst_ids))
        state_attr = torch.tensor([0.0, 0.0])
        broadcasted_state_feat = broadcast_states_to_bonds(g, state_attr)
        assert [broadcasted_state_feat.size(dim=0), broadcasted_state_feat.size(dim=1)] == [3, 2]
        broadcasted_state_feat = broadcast_states_to_atoms(g, state_attr)
        assert [broadcasted_state_feat.size(dim=0), broadcasted_state_feat.size(dim=1)] == [5, 2]

    def test_gaussian_expansion(self):
        bond_dist = torch.tensor([1.0])
        dist_converter = GaussianExpansion()
        expanded_dist = dist_converter(bond_dist)
        # check the shape of a vector
        assert np.allclose([expanded_dist.size(dim=0), expanded_dist.size(dim=1)], [1, 20])
        # check the first value of expanded distance
        assert np.allclose(expanded_dist[0][0], np.exp(-0.5 * np.power(1.0 - 0.0, 2.0)))
        # check the last value of expanded distance
        assert np.allclose(expanded_dist[0][-1], np.exp(-0.5 * np.power(1.0 - 4.0, 2.0)))


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import unittest

import numpy as np
import torch

from matgl.utils.maths import (
    GaussianExpansion,
    SphericalBesselFunction,
    SphericalHarmonicsFunction,
)


class TestGaussianAndSphericalBesselFuction(unittest.TestCase):
    def test_gaussian(self):
        r = torch.linspace(1.0, 5.0, 11)
        rbf_gaussian = GaussianExpansion(initial=0.0, final=5.0, num_centers=10, width=0.5)
        rbf = rbf_gaussian(r)
        assert [rbf.size(dim=0), rbf.size(dim=1)] == [11, 10]

    def test_sphericalbesselfunction(self):
        r = torch.linspace(1.0, 5.0, 11)
        rbf_sb = SphericalBesselFunction(max_n=3, max_l=3, cutoff=5.0, smooth=False)
        rbf = rbf_sb(r)
        assert [rbf.size(dim=0), rbf.size(dim=1)] == [11, 9]

        rbf_sb = SphericalBesselFunction(max_n=3, max_l=3, cutoff=5.0, smooth=True)
        rbf = rbf_sb(r)
        assert [rbf.size(dim=0), rbf.size(dim=1)] == [11, 3]

    def test_sphericalbesselfunction_smooth(self):
        r = torch.linspace(1.0, 5.0, 11)
        rbf_sb = SphericalBesselFunction(max_n=3, max_l=3, cutoff=5.0, smooth=False)
        rbf = rbf_sb(r)
        assert [rbf.size(dim=0), rbf.size(dim=1)] == [11, 9]

        rbf_sb = SphericalBesselFunction(max_n=3, max_l=3, cutoff=5.0, smooth=True)
        rbf = rbf_sb(r)
        assert [rbf.size(dim=0), rbf.size(dim=1)] == [11, 3]

    def test_spherical_harmonic_function(self):
        theta = torch.linspace(-1, 1, 10)
        phi = torch.linspace(0, 2 * np.pi, 10)
        abf_sb = SphericalHarmonicsFunction(max_l=3, use_phi=True)
        abf = abf_sb(theta, phi)
        assert [abf.size(dim=0), abf.size(dim=1)] == [10, 9]


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import os
import shutil
import unittest

import numpy as np
import pytorch_lightning as pl
from dgl.data.utils import split_dataset
from pymatgen.util.testing import PymatgenTest

from matgl.ext.pymatgen import Structure2Graph, get_element_list
from matgl.graph.data import M3GNetDataset, MEGNetDataset, MGLDataLoader, collate_fn, collate_fn_efs
from matgl.models import M3GNet, MEGNet
from matgl.utils.training import ModelTrainer, PotentialTrainer

module_dir = os.path.dirname(os.path.abspath(__file__))


class ModelTrainerTest(PymatgenTest):
    def test_megnet_training(self):
        s1 = self.get_structure("LiFePO4")
        s2 = self.get_structure("BaNiO3")
        structures = [s1, s1, s1, s1, s1, s1, s1, s1, s1, s1, s2, s2, s2, s2, s2, s2, s2, s2, s2, s2]
        label = np.zeros(20)
        element_types = get_element_list([s1, s2])
        cry_graph = Structure2Graph(element_types=element_types, cutoff=4.0)
        dataset = MEGNetDataset(structures=structures, converter=cry_graph, labels=label, label_name="label")
        train_data, val_data, test_data = split_dataset(
            dataset,
            frac_list=[0.8, 0.1, 0.1],
            shuffle=True,
            random_state=42,
        )
        train_loader, val_loader, test_loader = MGLDataLoader(
            train_data=train_data,
            val_data=val_data,
            test_data=test_data,
            collate_fn=collate_fn,
            batch_size=2,
            num_workers=1,
        )

        model = MEGNet(
            dim_node_embedding=16,
            dim_edge_embedding=100,
            dim_state_embedding=2,
            nblocks=3,
            hidden_layer_sizes_input=(64, 32),
            hidden_layer_sizes_conv=(64, 64, 32),
            nlayers_set2set=1,
            niters_set2set=3,
            hidden_layer_sizes_outputput=[32, 16],
            is_classification=False,
        )

        lit_model = ModelTrainer(model=model)
        trainer = pl.Trainer(max_epochs=2)
        trainer.fit(model=lit_model, train_dataloaders=train_loader, val_dataloaders=val_loader)

    def test_m3gnet_training(self):
        s1 = self.get_structure("LiFePO4")
        s2 = self.get_structure("BaNiO3")
        structures = [s1, s2, s1, s2, s1, s2, s1, s2, s1, s2, s1, s2, s1, s2, s1, s2, s1, s2, s1, s2]
        energies = np.zeros(20)
        f1 = np.zeros((28, 3)).tolist()
        f2 = np.zeros((10, 3)).tolist()
        s = np.zeros((3, 3)).tolist()
        forces = [f1, f2, f1, f2, f1, f2, f1, f2, f1, f2, f1, f2, f1, f2, f1, f2, f1, f2, f1, f2]
        stresses = [s, s, s, s, s, s, s, s, s, s, s, s, s, s, s, s, s, s, s, s]
        element_types = get_element_list([s1, s2])
        cry_graph = Structure2Graph(element_types=element_types, cutoff=5.0)
        dataset = M3GNetDataset(
            threebody_cutoff=4.0,
            structures=structures,
            converter=cry_graph,
            energies=energies,
            forces=forces,
            stresses=stresses,
        )
        train_data, val_data, test_data = split_dataset(
            dataset,
            frac_list=[0.8, 0.1, 0.1],
            shuffle=True,
            random_state=42,
        )
        train_loader, val_loader, test_loader = MGLDataLoader(
            train_data=train_data,
            val_data=val_data,
            test_data=test_data,
            collate_fn=collate_fn_efs,
            batch_size=2,
            num_workers=1,
        )
        model = M3GNet(
            element_types=element_types,
            is_intensive=False,
        )
        lit_model = PotentialTrainer(model=model)
        trainer = pl.Trainer(max_epochs=2)
        trainer.fit(model=lit_model, train_dataloaders=train_loader, val_dataloaders=val_loader)

    @classmethod
    def tearDownClass(cls):
        for fn in ("dgl_graph.bin", "dgl_line_graph.bin", "state_attr.pt", "labels.json"):
            try:
                os.remove(fn)
            except FileNotFoundError:
                pass

        shutil.rmtree("lightning_logs")


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import pytest
import torch

from matgl.data.transformer import Normalizer


class TestNormalizer:
    def test_transform(self):
        data = torch.randn(100) * 10 + 1
        transformer = Normalizer.from_data(data)
        scaled = transformer.transform(data)

        inverse = transformer.inverse_transform(scaled)
        assert inverse == pytest.approx(data, abs=1e-5)



from __future__ import annotations

import unittest

import numpy as np
import torch

from matgl.layers._activations import SoftExponential, SoftPlus2


class TestActivations(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.act1 = SoftPlus2()
        cls.act2 = SoftExponential()
        cls.act3 = SoftExponential(1.0)

        cls.x = torch.tensor([1.0, 2.0])

    def test_softplus2(self):
        out = self.act1(self.x)
        np.testing.assert_allclose(out.numpy(), np.array([0.62011445, 1.4337809]))

    def test_soft_exponential(self):
        out = self.act2(self.x)
        np.testing.assert_allclose(out.numpy(), np.array([1.0, 2.0]))
        out = self.act3(self.x)
        np.testing.assert_allclose(out.detach().numpy(), np.array([2.7182817, 7.389056]))


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import shutil
import unittest

import torch

from matgl.utils.io import RemoteFile


class ModelSourceTestCase(unittest.TestCase):
    def test_remote(self):
        with RemoteFile(
            "https://github.com/materialsvirtuallab/matgl/raw/main/pretrained_models/MEGNet-MP-2018.6.1-Eform/model.pt",
            cache_location=".",
        ) as s:
            d = torch.load(s, map_location=torch.device("cpu"))
            assert "nblocks" in d["model"]["init_args"]
        try:  # cleanup
            shutil.rmtree("MEGNet-MP-2018.6.1-Eform")
        except FileNotFoundError:
            pass


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import os
import unittest

import torch as th
from pymatgen.core.structure import Lattice, Structure

from matgl.ext.pymatgen import Structure2Graph, get_element_list
from matgl.graph.compute import compute_pair_vector_and_distance
from matgl.layers import BondExpansion
from matgl.models import MEGNet


class MEGNetTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        s = Structure(Lattice.cubic(4.0), ["Mo", "S"], [[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]])

        cls.element_types = get_element_list([s])  # type: ignore
        p2g = Structure2Graph(element_types=cls.element_types, cutoff=5.0)  # type: ignore
        graph, state = p2g.get_graph(s)
        cls.g1 = graph  # type: ignore
        cls.state1 = state  # type: ignore

    def test_megnet(self):
        model = MEGNet(
            dim_node_embedding=16,
            dim_edge_embedding=100,
            dim_state_embedding=2,
            nblocks=3,
            include_states=True,
            hidden_layer_sizes_input=(64, 32),
            hidden_layer_sizes_conv=(64, 64, 32),
            activation_type="swish",
            nlayers_set2set=4,
            niters_set2set=3,
            hidden_layer_sizes_output=(32, 16),
            is_classification=True,
        )
        bond_expansion = BondExpansion(rbf_type="Gaussian", initial=0.0, final=6.0, num_centers=100, width=0.5)
        bond_vec, bond_dist = compute_pair_vector_and_distance(self.g1)
        self.g1.edata["edge_attr"] = bond_expansion(bond_dist)
        self.state1 = th.tensor(self.state1)
        output = model(self.g1, self.g1.edata["edge_attr"], self.g1.ndata["node_type"], self.state1)
        assert [th.numel(output)] == [1]

    def test_save_load(self):
        model = MEGNet(
            dim_node_embedding=16,
            dim_edge_embedding=100,
            dim_state_embedding=2,
            nblocks=3,
            include_states=True,
            hidden_layer_sizes_input=(64, 32),
            hidden_layer_sizes_conv=(64, 64, 32),
            activation_type="swish",
            nlayers_set2set=4,
            niters_set2set=3,
            hidden_layer_sizes_output=(32, 16),
            is_classification=True,
        )
        model.save(".", metadata={"description": "forme model"})
        MEGNet.load(".")
        os.remove("model.pt")
        os.remove("model.json")
        os.remove("state.pt")


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import unittest

import torch
from pymatgen.core.structure import Lattice, Structure
from torch import nn

from matgl.ext.pymatgen import Structure2Graph, get_element_list
from matgl.graph.compute import compute_pair_vector_and_distance
from matgl.layers import BondExpansion, EmbeddingBlock
from matgl.layers._core import MLP, GatedMLP


class TestCoreAndEmbedding(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.s1 = Structure(Lattice.cubic(4.0), ["Mo", "Mo"], [[0.0, 0, 0], [0.5, 0.5, 0.5]])
        Structure(Lattice.cubic(3), ["Mo", "Fe"], [[0, 0, 0], [0.5, 0.5, 0.5]])
        element_types = get_element_list([cls.s1])
        p2g = Structure2Graph(element_types=element_types, cutoff=4.0)
        graph, state = p2g.get_graph(cls.s1)
        cls.g1 = graph
        cls.state1 = state

        bond_vec, bond_dist = compute_pair_vector_and_distance(cls.g1)
        cls.g1.edata["bond_dist"] = bond_dist

        cls.x = torch.randn(4, 10, requires_grad=True)

    def test_mlp(self):
        layer = MLP(dims=[10, 3], activation=nn.SiLU())
        out = layer(self.x).double()
        assert [out.size()[0], out.size()[1]] == [4, 3]

    def test_gated_mlp(self):
        torch.manual_seed(42)
        layer = GatedMLP(in_feats=10, dims=[10, 1], activate_last=False)
        out = layer(self.x)
        assert [out.size()[0], out.size()[1]] == [4, 1]

    def test_embedding(self):
        bond_expansion = BondExpansion(rbf_type="SphericalBessel", max_n=3, max_l=3, cutoff=4.0, smooth=False)
        bond_dist = self.g1.edata["bond_dist"]
        bond_basis = bond_expansion(bond_dist)
        # include state features
        embed = EmbeddingBlock(
            degree_rbf=9,
            dim_node_embedding=16,
            dim_edge_embedding=16,
            dim_state_feats=16,
            include_state=True,
            activation=nn.SiLU(),
        )
        state_attr = torch.tensor([1.0, 2.0])
        node_attr = self.g1.ndata["attr"]
        edge_attr = bond_basis
        node_feat, edge_feat, state_feat = embed(node_attr, edge_attr, state_attr)

        assert [node_feat.size(dim=0), node_feat.size(dim=1)] == [2, 16]
        assert [edge_feat.size(dim=0), edge_feat.size(dim=1)] == [28, 16]
        assert [state_feat.size(dim=0), state_feat.size(dim=1)] == [1, 16]
        # include state embedding
        embed2 = EmbeddingBlock(
            degree_rbf=9,
            dim_node_embedding=16,
            dim_edge_embedding=16,
            include_state=True,
            dim_state_embedding=32,
            ntypes_state=2,
            activation=nn.SiLU(),
        )
        node_feat, edge_feat, state_feat = embed2(node_attr, edge_attr, torch.tensor([1]))
        assert [state_feat.size(dim=0), state_feat.size(dim=1)] == [1, 32]
        # include state features
        embed3 = EmbeddingBlock(
            degree_rbf=9,
            dim_node_embedding=16,
            dim_edge_embedding=16,
            dim_state_feats=16,
            include_state=True,
            activation=nn.SiLU(),
        )
        node_feat, edge_feat, state_feat = embed3(node_attr, edge_attr, torch.tensor([1.0, 2.0]))
        assert [state_feat.size(dim=0), state_feat.size(dim=1)] == [1, 16]
        # without any state feature
        embed4 = EmbeddingBlock(degree_rbf=9, dim_node_embedding=16, dim_edge_embedding=16, activation=nn.SiLU())
        node_feat, edge_feat, state_feat = embed4(
            node_attr, edge_attr, torch.tensor([0.0, 0.0])
        )  # this will be default value
        assert state_feat is None


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import unittest

import numpy as np
import torch
from pymatgen.core import Lattice, Structure

from matgl.apps.pes import Potential
from matgl.ext.pymatgen import Structure2Graph, get_element_list
from matgl.models._m3gnet import M3GNet


class TestPotential(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        s = Structure(Lattice.cubic(4.0), ["Mo", "S"], [[0.0025, 0.0, 0.0], [0.5, 0.5, 0.5]])
        s.states = np.array([[0.1, 0.2, 0.3, 0.4, 0.5]])

        cls.element_types = get_element_list([s])
        p2g = Structure2Graph(element_types=cls.element_types, cutoff=5.0)
        graph, state = p2g.get_graph(s)
        cls.g1 = graph
        cls.state1 = state

    def test_potential_efsh(self):
        model = M3GNet(element_types=self.element_types, is_intensive=False)
        ff = Potential(model=model, calc_hessian=True)
        e, f, s, h = ff(self.g1, self.state1)
        assert [torch.numel(e)] == [1]
        assert [f.size(dim=0), f.size(dim=1)] == [2, 3]
        assert [s.size(dim=0), s.size(dim=1)] == [3, 3]
        assert [h.size(dim=0), h.size(dim=1)] == [6, 6]

    def test_potential_efs(self):
        model = M3GNet(element_types=self.element_types, is_intensive=False)
        ff = Potential(model=model)
        e, f, s, h = ff(self.g1, self.state1)
        assert [torch.numel(e)] == [1]
        assert [f.size(dim=0), f.size(dim=1)] == [2, 3]
        assert [s.size(dim=0), s.size(dim=1)] == [3, 3]
        assert [h.size(dim=0)] == [1]

    def test_potential_ef(self):
        model = M3GNet(element_types=self.element_types, is_intensive=False)
        ff = Potential(model=model, calc_stresses=False)
        e, f, s, h = ff(self.g1, self.state1)
        assert [torch.numel(e)] == [1]
        assert [f.size(dim=0), f.size(dim=1)] == [2, 3]
        assert [s.size(dim=0)] == [1]
        assert [h.size(dim=0)] == [1]

    def test_potential_e(self):
        model = M3GNet(element_types=self.element_types, is_intensive=False)
        ff = Potential(model=model, calc_forces=False, calc_stresses=False)
        e, f, s, h = ff(self.g1, self.state1)
        assert [torch.numel(e)] == [1]
        assert [f.size(dim=0)] == [1]
        assert [s.size(dim=0)] == [1]
        assert [h.size(dim=0)] == [1]


if __name__ == "__main__":
    unittest.main()



"""Global configuration variables for matgl."""
from __future__ import annotations

import os
from pathlib import Path

import numpy as np
import torch

DEFAULT_ELEMENT_TYPES = (
    "H",
    "He",
    "Li",
    "Be",
    "B",
    "C",
    "N",
    "O",
    "F",
    "Ne",
    "Na",
    "Mg",
    "Al",
    "Si",
    "P",
    "S",
    "Cl",
    "Ar",
    "K",
    "Ca",
    "Sc",
    "Ti",
    "V",
    "Cr",
    "Mn",
    "Fe",
    "Co",
    "Ni",
    "Cu",
    "Zn",
    "Ga",
    "Ge",
    "As",
    "Se",
    "Br",
    "Kr",
    "Rb",
    "Sr",
    "Y",
    "Zr",
    "Nb",
    "Mo",
    "Tc",
    "Ru",
    "Rh",
    "Pd",
    "Ag",
    "Cd",
    "In",
    "Sn",
    "Sb",
    "Te",
    "I",
    "Xe",
    "Cs",
    "Ba",
    "La",
    "Ce",
    "Pr",
    "Nd",
    "Pm",
    "Sm",
    "Eu",
    "Gd",
    "Tb",
    "Dy",
    "Ho",
    "Er",
    "Tm",
    "Yb",
    "Lu",
    "Hf",
    "Ta",
    "W",
    "Re",
    "Os",
    "Ir",
    "Pt",
    "Au",
    "Hg",
    "Tl",
    "Pb",
    "Bi",
    "Ac",
    "Th",
    "Pa",
    "U",
    "Np",
    "Pu",
)

DTYPES = {
    "float32": {"numpy": np.float32, "torch": torch.float32},
    "float16": {"numpy": np.float16, "torch": torch.float16},
    "int32": {"numpy": np.int32, "torch": torch.int32},
    "int16": {"numpy": np.int16, "torch": torch.int16},
}

MATGL_CACHE = Path(os.path.expanduser("~")) / ".matgl"
PRETRAINED_MODELS_BASE_URL = "https://github.com/materialsvirtuallab/matgl/raw/main/pretrained_models/"



from __future__ import annotations

import unittest

import torch
from pymatgen.core.structure import Lattice, Structure
from torch import nn

from matgl.ext.pymatgen import Structure2Graph, get_element_list
from matgl.graph.compute import (
    compute_pair_vector_and_distance,
    compute_theta_and_phi,
    create_line_graph,
)
from matgl.layers import BondExpansion, EmbeddingBlock
from matgl.layers._core import MLP, GatedMLP
from matgl.layers._three_body import SphericalBesselWithHarmonics, ThreeBodyInteractions
from matgl.utils.cutoff import polynomial_cutoff


class TestThreeBody(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.s = Structure(Lattice.cubic(4.0), ["Mo", "S"], [[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]])

        element_types = get_element_list([cls.s])
        p2g = Structure2Graph(element_types=element_types, cutoff=5.0)
        graph, state = p2g.get_graph(cls.s)
        cls.g1 = graph
        cls.state1 = state
        bond_vec, bond_dist = compute_pair_vector_and_distance(cls.g1)
        cls.g1.edata["bond_dist"] = bond_dist
        cls.g1.edata["bond_vec"] = bond_vec

    def test_spherical_bessel_with_harmonics(self):
        sb_and_sh = SphericalBesselWithHarmonics(max_n=3, max_l=3, cutoff=5.0, use_smooth=False, use_phi=False)
        l_g1 = create_line_graph(self.g1, threebody_cutoff=4.0)
        l_g1.apply_edges(compute_theta_and_phi)
        three_body_basis = sb_and_sh(l_g1)
        assert [three_body_basis.size(dim=0), three_body_basis.size(dim=1)] == [364, 9]

        sb_and_sh = SphericalBesselWithHarmonics(max_n=3, max_l=2, cutoff=5.0, use_smooth=False, use_phi=True)
        l_g1 = create_line_graph(self.g1, threebody_cutoff=4.0)
        l_g1.apply_edges(compute_theta_and_phi)
        three_body_basis = sb_and_sh(l_g1)
        assert [three_body_basis.size(dim=0), three_body_basis.size(dim=1)] == [364, 12]

        sb_and_sh = SphericalBesselWithHarmonics(max_n=3, max_l=3, cutoff=5.0, use_smooth=True, use_phi=False)
        l_g1 = create_line_graph(self.g1, threebody_cutoff=4.0)
        l_g1.apply_edges(compute_theta_and_phi)
        three_body_basis = sb_and_sh(l_g1)
        assert [three_body_basis.size(dim=0), three_body_basis.size(dim=1)] == [364, 9]

        sb_and_sh = SphericalBesselWithHarmonics(max_n=3, max_l=3, cutoff=5.0, use_smooth=True, use_phi=True)
        l_g1 = create_line_graph(self.g1, threebody_cutoff=4.0)
        l_g1.apply_edges(compute_theta_and_phi)
        three_body_basis = sb_and_sh(l_g1)
        assert [three_body_basis.size(dim=0), three_body_basis.size(dim=1)] == [364, 27]

    def test_three_body_interactions(self):
        l_g1 = create_line_graph(self.g1, threebody_cutoff=4.0)
        l_g1.apply_edges(compute_theta_and_phi)
        bond_expansion = BondExpansion(max_l=3, max_n=3, cutoff=5.0, rbf_type="SphericalBessel", smooth=False)
        bond_basis = bond_expansion(self.g1.edata["bond_dist"])
        self.g1.edata["rbf"] = bond_basis
        sb_and_sh = SphericalBesselWithHarmonics(max_n=3, max_l=3, cutoff=5.0, use_smooth=False, use_phi=False)
        three_body_basis = sb_and_sh(l_g1)
        three_body_cutoff = polynomial_cutoff(self.g1.edata["bond_dist"], 4.0)
        max_n = 3
        max_l = 3
        num_node_feats = 16
        num_edge_feats = 16
        state_attr = torch.tensor([0.0, 0.0])
        embedding = EmbeddingBlock(
            degree_rbf=9, dim_node_embedding=num_node_feats, dim_edge_embedding=num_edge_feats, activation=nn.SiLU()
        )

        node_attr = self.g1.ndata["attr"]
        edge_attr = self.g1.edata["rbf"]
        node_feat, edge_feat, state_feat = embedding(node_attr, edge_attr, state_attr)
        degree = max_n * max_l
        three_body_interactions = ThreeBodyInteractions(
            update_network_atom=MLP(dims=[num_node_feats, degree], activation=nn.Sigmoid(), activate_last=True),
            update_network_bond=GatedMLP(in_feats=degree, dims=[num_edge_feats], use_bias=False),
        )
        edge_feat_updated = three_body_interactions(
            self.g1, l_g1, three_body_basis, three_body_cutoff, node_feat, edge_feat
        )
        assert [edge_feat_updated.size(dim=0), edge_feat_updated.size(dim=1)] == [28, 16]


if __name__ == "__main__":
    unittest.main()



"""
Embedding node, edge and optional state attributes
"""
from __future__ import annotations

import torch
from torch import nn

from matgl.layers._core import MLP


class EmbeddingBlock(nn.Module):
    """
    Embedding block for generating node, bond and state features
    """

    def __init__(
        self,
        degree_rbf: int,
        activation: nn.Module,
        dim_node_embedding: int,
        dim_edge_embedding: int | None = None,
        dim_state_feats: int | None = None,
        ntypes_node: int | None = None,
        include_state: bool = False,
        ntypes_state: int | None = None,
        dim_state_embedding: int | None = None,
    ):
        """

        Args:
            degree_rbf (int): number of rbf
            activation (nn.Module): activation type
            dim_node_embedding (int): dimensionality of node features
            dim_edge_embedding (int): dimensionality of edge features
            dim_state_feats: dimensionality of state features
            ntypes_node: number of node labels
            include_state: Whether to include state embedding
            ntypes_state: number of state labels
            dim_state_embedding: dimensionality of state embedding
        """
        super().__init__()
        self.include_state = include_state
        self.ntypes_state = ntypes_state
        self.dim_node_embedding = dim_node_embedding
        self.dim_edge_embedding = dim_edge_embedding
        self.dim_state_feats = dim_state_feats
        self.ntypes_node = ntypes_node
        self.dim_state_embedding = dim_state_embedding
        self.activation = activation
        if ntypes_state and dim_state_embedding is not None:
            self.layer_state_embedding = nn.Embedding(ntypes_state, dim_state_embedding)  # type: ignore
        if ntypes_node is not None:
            self.layer_node_embedding = nn.Embedding(ntypes_node, dim_node_embedding)
        if dim_edge_embedding is not None:
            dim_edges = [degree_rbf, dim_edge_embedding]
            self.layer_edge_embedding = MLP(dim_edges, activation=activation, activate_last=True)

    def forward(self, node_attr, edge_attr, state_attr):
        """
        Output embedded features

        Args:
        node_attr: node attribute
        edge_attr: edge attribute
        state_attr: state attribute

        Returns:
        node_feat: embedded node features
        edge_feat: embedded edge features
        state_feat: embedded state features
        """
        if self.ntypes_node is not None:
            node_feat = self.layer_node_embedding(node_attr)
        else:
            node_embed = MLP([node_attr.shape[-1], self.dim_node_embedding], activation=self.activation)
            node_feat = node_embed(node_attr.to(torch.float32))
        if self.dim_edge_embedding is not None:
            edge_feat = self.layer_edge_embedding(edge_attr.to(torch.float32))
        else:
            edge_feat = edge_attr
        if self.include_state is True:
            if self.ntypes_state and self.dim_state_embedding is not None:
                state_feat = self.layer_state_embedding(state_attr)
            elif self.dim_state_feats is not None:
                state_attr = torch.unsqueeze(state_attr, 0)
                state_embed = MLP([state_attr.shape[-1], self.dim_state_feats], activation=self.activation)
                state_feat = state_embed(state_attr.to(torch.float32))
            else:
                state_feat = state_attr
        else:
            state_feat = None
        return node_feat, edge_feat, state_feat



from __future__ import annotations

import unittest

import pytest
import torch.nn

from matgl.data.transformer import Normalizer
from matgl.models._wrappers import TransformedTargetModel


class DummyModel(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self):
        return torch.ones(1)


class TransformedTargetModelTest(unittest.TestCase):
    def test_forward(self):
        model = TransformedTargetModel(DummyModel(), Normalizer(1, 2))
        assert float(model.forward()), pytest.approx(3)


if __name__ == "__main__":
    unittest.main()



"""
Tools to construct a dataset of DGL graphs
"""
from __future__ import annotations

import json
import os
from typing import TYPE_CHECKING, Callable

import dgl
import numpy as np
import torch
from dgl.data import DGLDataset
from dgl.data.utils import load_graphs, save_graphs
from dgl.dataloading import GraphDataLoader
from tqdm import trange

from matgl.graph.compute import compute_pair_vector_and_distance, create_line_graph
from matgl.layers import BondExpansion

if TYPE_CHECKING:
    from matgl.graph.converters import GraphConverter


def collate_fn(batch):
    """
    Merge a list of dgl graphs to form a batch
    """
    graphs, labels, state_attr = map(list, zip(*batch))
    g = dgl.batch(graphs)
    labels = torch.tensor(labels, dtype=torch.float32)
    state_attr = torch.stack(state_attr)
    return g, labels, state_attr


def collate_fn_efs(batch):
    """
    Merge a list of dgl graphs to form a batch
    """
    graphs, line_graphs, state_attr, energies, forces, stresses = map(list, zip(*batch))
    g = dgl.batch(graphs)
    l_g = dgl.batch(line_graphs)
    e = torch.tensor(energies, dtype=torch.float32)
    f = torch.vstack(forces)
    s = torch.vstack(stresses)
    state_attr = torch.stack(state_attr)
    return g, l_g, state_attr, e, f, s


def MGLDataLoader(
    train_data: dgl.data.utils.Subset,
    val_data: dgl.data.utils.Subset,
    collate_fn: Callable,
    batch_size: int,
    num_workers: int,
    use_ddp: bool = False,
    pin_memory: bool = False,
    test_data: dgl.data.utils.Subset | None = None,
    generator: torch.Generator | None = None,
):
    """
    Dataloader for MEGNet training
    Args:
    train_data: training data
    val_data: validation data
    test_data: testing data
    collate_fn:
    """
    train_loader = GraphDataLoader(
        train_data,
        batch_size=batch_size,
        shuffle=True,
        collate_fn=collate_fn,
        num_workers=num_workers,
        pin_memory=pin_memory,
        use_ddp=use_ddp,
        generator=generator,
    )

    val_loader = GraphDataLoader(
        val_data,
        batch_size=batch_size,
        shuffle=False,
        collate_fn=collate_fn,
        num_workers=num_workers,
        pin_memory=pin_memory,
    )
    if test_data is not None:
        test_loader = GraphDataLoader(
            test_data,
            batch_size=batch_size,
            shuffle=False,
            collate_fn=collate_fn,
            num_workers=num_workers,
            pin_memory=pin_memory,
        )
        return train_loader, val_loader, test_loader
    return train_loader, val_loader


class MEGNetDataset(DGLDataset):
    """
    Create a dataset including dgl graphs
    """

    def __init__(
        self,
        structures: list,
        labels: list,
        label_name: str,
        converter: GraphConverter,
        initial: float = 0.0,
        final: float = 5.0,
        num_centers: int = 100,
        width: float = 0.5,
        name: str = "MEGNETDataset",
        graph_labels: list | None = None,
    ):
        """
        Args:
        structures: Pymatgen strutcure
        labels: property values
        label_name: label name
        converter: Transformer for converting structures to DGL graphs, e.g., Pmg2Graph.
        initial: initial distance for Gaussian expansions
        final: final distance for Gaussian expansions
        num_centers: number of Gaussian functions
        width: width of Gaussian functions
        name: Name of dataset
        graph_labels: graph attributes either integers and floating point numbers
        """
        self.converter = converter
        self.structures = structures
        self.labels = torch.FloatTensor(labels)
        self.label_name = label_name
        self.initial = initial
        self.final = final
        self.num_centers = num_centers
        self.width = width
        self.graph_labels = graph_labels

        super().__init__(name=name)

    def has_cache(self, filename: str = "dgl_graph.bin") -> bool:
        """
        Check if the dgl_graph.bin exists or not
        Args:
            :filename: Name of file storing dgl graphs
        Returns: True if file exists.
        """
        return os.path.exists(filename)

    def process(self) -> tuple:
        """
        Convert Pymatgen structure into dgl graphs
        """
        num_graphs = self.labels.shape[0]
        self.graphs = []
        self.state_attr = []
        bond_expansion = BondExpansion(
            rbf_type="Gaussian", initial=self.initial, final=self.final, num_centers=self.num_centers, width=self.width
        )
        for idx in trange(num_graphs):
            structure = self.structures[idx]
            graph, state_attr = self.converter.get_graph(structure)
            bond_vec, bond_dist = compute_pair_vector_and_distance(graph)
            graph.edata["edge_attr"] = bond_expansion(bond_dist)
            self.graphs.append(graph)
            self.state_attr.append(state_attr)
        if self.graph_labels is not None:
            if np.array(self.graph_labels).dtype == "int64":
                self.state_attr = torch.tensor(self.graph_labels).long()  # type: ignore
            else:
                self.state_attr = torch.tensor(self.graph_labels)  # type: ignore
        else:
            self.state_attr = torch.tensor(self.state_attr)  # type: ignore

        return self.graphs, self.state_attr

    def save(self, filename: str = "dgl_graph.bin", filename_state_attr: str = "state_attr.pt"):
        """
        Save dgl graphs
        Args:
        :filename: Name of file storing dgl graphs
        :filename_state_attr: Name of file storing graph attrs
        """
        labels_with_key = {self.label_name: self.labels}
        save_graphs(filename, self.graphs, labels_with_key)
        torch.save(self.state_attr, filename_state_attr)

    def load(self, filename: str = "dgl_graph.bin", filename_state_attr: str = "state_attr.pt"):
        """
        Load dgl graphs
        Args:
        :filename: Name of file storing dgl graphs
        :filename: Name of file storing state attrs
        """
        self.graphs, label_dict = load_graphs(filename)
        self.label = torch.stack([label_dict[key] for key in self.label_keys], dim=1)
        self.state_attr = torch.load("state_attr.pt")

    def __getitem__(self, idx: int):
        """
        Get graph and label with idx
        """
        return self.graphs[idx], self.labels[idx], self.state_attr[idx]

    def __len__(self):
        """
        Get size of dataset
        """
        return len(self.graphs)


class M3GNetDataset(DGLDataset):
    """
    Create a dataset including dgl graphs
    """

    def __init__(
        self,
        structures: list,
        energies: list,
        forces: list,
        stresses: None | list,
        converter: GraphConverter,
        threebody_cutoff: float,
        name="M3GNETDataset",
        graph_labels: list | None = None,
    ):
        """
        Args:
        structures: Pymatgen strutcure
        energies: Target energies
        forces: Target forces
        stresses: Target stresses
        converter: dgl graph converter
        threebody_cutoff: cutoff for three body
        name: name of dataset
        graph_labels: state attributes
        """
        self.converter = converter
        self.structures = structures
        self.energies = energies
        self.forces = forces
        self.threebody_cutoff = threebody_cutoff
        self.stresses = np.zeros(len(self.energies)) if stresses is None else stresses
        self.graph_labels = graph_labels
        super().__init__(name=name)

    def has_cache(self, filename: str = "dgl_graph.bin") -> bool:
        """
        Check if the dgl_graph.bin exists or not
        Args:
            :filename: Name of file storing dgl graphs
        Returns: True if file exists.
        """
        return os.path.exists(filename)

    def process(self) -> tuple:
        """
        Convert Pymatgen structure into dgl graphs
        """
        num_graphs = len(self.energies)
        self.graphs = []
        self.line_graphs = []
        self.state_attr = []
        for idx in trange(num_graphs):
            structure = self.structures[idx]
            graph, state_attr = self.converter.get_graph(structure)
            self.graphs.append(graph)
            self.state_attr.append(state_attr)
            bond_vec, bond_dist = compute_pair_vector_and_distance(graph)
            graph.edata["bond_vec"] = bond_vec
            graph.edata["bond_dist"] = bond_dist
            line_graph = create_line_graph(graph, self.threebody_cutoff)
            line_graph.ndata.pop("bond_vec")
            line_graph.ndata.pop("bond_dist")
            line_graph.ndata.pop("pbc_offset")
            self.line_graphs.append(line_graph)
        if self.graph_labels is not None:
            self.state_attr = torch.tensor(self.graph_labels).long()  # type: ignore
        else:
            self.state_attr = torch.tensor(self.state_attr)  # type: ignore

        return self.graphs, self.line_graphs, self.state_attr

    def save(
        self,
        filename: str = "dgl_graph.bin",
        filename_line_graph: str = "dgl_line_graph.bin",
        filename_state_attr: str = "state_attr.pt",
    ):
        """
        Save dgl graphs
        Args:
        :filename: Name of file storing dgl graphs
        :filename_state_attr: Name of file storing graph attrs
        """
        labels_with_key = {"energies": self.energies, "forces": self.forces, "stresses": self.stresses}
        save_graphs(filename, self.graphs)
        save_graphs(filename_line_graph, self.line_graphs)
        torch.save(self.state_attr, filename_state_attr)
        with open("labels.json", "w") as file:
            file.write("".join(str(labels_with_key).split("\n")))

    def load(
        self,
        filename: str = "dgl_graph.bin",
        filename_line_graph: str = "dgl_line_graph.bin",
        filename_state_attr: str = "state_attr.pt",
    ):
        """
        Load dgl graphs
        Args:
        :filename: Name of file storing dgl graphs
        :filename: Name of file storing state attrs
        """
        self.graphs = load_graphs(filename)
        self.line_graphs = load_graphs(filename_line_graph)
        with open("labels.json") as file:
            labels = json.load(file)
        self.energies = labels["energies"]
        self.forces = labels["forces"]
        self.stresses = labels["stresses"]
        self.state_attr = torch.load("state_attr.pt")

    def __getitem__(self, idx: int):
        """
        Get graph and label with idx
        """
        return (
            self.graphs[idx],
            self.line_graphs[idx],
            self.state_attr[idx],
            self.energies[idx],
            torch.tensor(self.forces[idx]),
            torch.tensor(self.stresses[idx]),
        )

    def __len__(self):
        """
        Get size of dataset
        """
        return len(self.graphs)



from __future__ import annotations

import unittest

import dgl
import numpy as np
import torch
from pymatgen.core.structure import Lattice, Structure

from matgl.ext.pymatgen import Structure2Graph, get_element_list
from matgl.layers._atom_ref import AtomRef


class TestAtomRef(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.s1 = Structure(Lattice.cubic(3.17), ["Mo", "S", "H"], [[0, 0, 0], [0.5, 0.5, 0.5], [0.75, 0.75, 0.75]])
        cls.element_types = get_element_list([cls.s1])
        p2g = Structure2Graph(element_types=cls.element_types, cutoff=4.0)
        graph, state = p2g.get_graph(cls.s1)
        cls.g1 = graph
        cls.state1 = state

    def test_atom_ref(self):
        element_ref = AtomRef(np.array([0.5, 1.0, 2.0]))

        atom_ref = element_ref(self.g1)
        assert atom_ref == 3.5

    def test_atom_ref_fit(self):
        element_ref = AtomRef(np.array([0.5, 1.0, 2.0]))
        properties = torch.tensor([2.0, 2.0])
        bg = dgl.batch([self.g1, self.g1])
        element_ref.fit([self.g1, self.g1], self.element_types, properties)
        atom_ref = element_ref(bg)
        assert list(np.round(atom_ref.numpy())) == [2.0, 2.0]

    def test_atom_ref_with_states(self):
        element_ref = AtomRef(np.array([[0.5, 1.0, 2.0], [2.0, 3.0, 5.0]]))
        state_label = torch.tensor([1])
        atom_ref = element_ref(self.g1, state_label)
        assert atom_ref == 10


if __name__ == "__main__":
    unittest.main()



"""
Tools to convert materials representations from Pymatgen and other codes to DGLGraphs.
"""
from __future__ import annotations

import abc

import dgl


class GraphConverter(metaclass=abc.ABCMeta):
    """
    Abstract base class for converters from input structures/molecules to graphs.
    """

    @abc.abstractmethod
    def get_graph(self, structure) -> tuple[dgl.DGLGraph, list]:
        """

        Args:
            structure: Input structure or molecule.

        Returns:
            DGLGraph object, state_attr
        """



"""
Interface with pymatgen objects.
"""
from __future__ import annotations

import dgl
import numpy as np
import scipy.sparse as sp
import torch
from dgl.backend import tensor
from pymatgen.core import Element, Molecule, Structure
from pymatgen.optimization.neighbors import find_points_in_spheres

from matgl.graph.converters import GraphConverter


def get_element_list(train_structures: list[Structure | Molecule]) -> tuple[str]:
    """Get the dictionary containing elements in the training set for atomic features

    Args:
        train_structures: pymatgen Molecule/Structure object

    Returns:
        Tuple of elements covered in training set
    """
    elements = set()
    for s in train_structures:
        elements.update(s.composition.get_el_amt_dict().keys())
    return tuple(sorted(elements, key=lambda el: Element(el).Z))  # type: ignore


class Molecule2Graph(GraphConverter):
    """
    Construct a DGL graph from Pymatgen Molecules.
    """

    def __init__(
        self,
        element_types: tuple[str, ...],
        cutoff: float = 5.0,
    ):
        """
        Parameters
        ----------
        element_types: List of elements present in dataset for graph conversion. This ensures all graphs are
            constructed with the same dimensionality of features.
        cutoff: Cutoff radius for graph representation
        """
        self.element_types = tuple(element_types)
        self.cutoff = cutoff

    def get_graph(self, mol: Molecule) -> tuple[dgl.DGLGraph, list]:
        """
        Get a DGL graph from an input molecule.

        :param mol: pymatgen molecule object
        :return: (dgl graph, state features)
        """
        natoms = len(mol)
        R = mol.cart_coords
        element_types = self.element_types
        Z = np.array([np.eye(len(element_types))[element_types.index(site.specie.symbol)] for site in mol])
        np.array([site.specie.Z for site in mol])
        weight = mol.composition.weight / len(mol)
        dist = np.linalg.norm(R[:, None, :] - R[None, :, :], axis=-1)
        dists = mol.distance_matrix.flatten()
        nbonds = (np.count_nonzero(dists <= self.cutoff) - natoms) / 2
        nbonds /= natoms
        adj = sp.csr_matrix(dist <= self.cutoff) - sp.eye(natoms, dtype=np.bool_)
        adj = adj.tocoo()
        u, v = tensor(adj.row), tensor(adj.col)
        g = dgl.graph((u, v))
        g = dgl.to_bidirected(g)
        g.ndata["pos"] = tensor(R)
        g.ndata["attr"] = tensor(Z)
        g.ndata["node_type"] = tensor(np.hstack([[element_types.index(site.specie.symbol)] for site in mol]))
        g.edata["pbc_offset"] = torch.zeros(g.num_edges(), 3)
        g.edata["lattice"] = torch.zeros(g.num_edges(), 3, 3)
        state_attr = [weight, nbonds]
        g.edata["pbc_offshift"] = torch.zeros(g.num_edges(), 3)

        return g, state_attr


class Structure2Graph(GraphConverter):
    """
    Construct a DGL graph from Pymatgen Structure.
    """

    def __init__(
        self,
        element_types: tuple[str, ...],
        cutoff: float = 5.0,
    ):
        """
        Parameters
        ----------
        element_types: List of elements present in dataset for graph conversion. This ensures all graphs are
            constructed with the same dimensionality of features.
        cutoff: Cutoff radius for graph representation
        """
        self.element_types = tuple(element_types)
        self.cutoff = cutoff

    def get_graph(self, structure: Structure) -> tuple[dgl.DGLGraph, list]:
        """
        Get a DGL graph from an input Structure.

        :param structure: pymatgen structure object
        :return:
            g: DGL graph
            state_attr: state features
        """
        numerical_tol = 1.0e-8
        pbc = np.array([1, 1, 1], dtype=int)
        element_types = self.element_types
        Z = np.array([np.eye(len(element_types))[element_types.index(site.specie.symbol)] for site in structure])
        atomic_number = np.array([site.specie.Z for site in structure])
        lattice_matrix = np.ascontiguousarray(np.array(structure.lattice.matrix), dtype=float)
        volume = structure.volume
        cart_coords = np.ascontiguousarray(np.array(structure.cart_coords), dtype=float)
        src_id, dst_id, images, bond_dist = find_points_in_spheres(
            cart_coords,
            cart_coords,
            r=self.cutoff,
            pbc=pbc,
            lattice=lattice_matrix,
            tol=numerical_tol,
        )
        exclude_self = (src_id != dst_id) | (bond_dist > numerical_tol)
        src_id, dst_id, images, bond_dist = (
            src_id[exclude_self],
            dst_id[exclude_self],
            images[exclude_self],
            bond_dist[exclude_self],
        )
        u, v = tensor(src_id), tensor(dst_id)
        g = dgl.graph((u, v))
        g.edata["pbc_offset"] = torch.tensor(images)
        g.edata["lattice"] = tensor([[lattice_matrix] for i in range(g.num_edges())])
        g.ndata["attr"] = tensor(Z)
        g.ndata["node_type"] = tensor(np.hstack([[element_types.index(site.specie.symbol)] for site in structure]))
        g.ndata["pos"] = tensor(cart_coords)
        g.ndata["volume"] = tensor([volume for i in range(atomic_number.shape[0])])
        state_attr = [0.0, 0.0]
        return g, state_attr



from __future__ import annotations

import os
import unittest

import numpy as np
from pymatgen.core import Lattice, Molecule, Structure
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.util.testing import PymatgenTest

from matgl.ext.ase import Atoms2Graph
from matgl.ext.pymatgen import Molecule2Graph, Structure2Graph, get_element_list

module_dir = os.path.dirname(os.path.abspath(__file__))


class Pmg2GraphTest(PymatgenTest):
    def test_get_graph_from_molecule(self):
        coords = [
            [0.000000, 0.000000, 0.000000],
            [0.000000, 0.000000, 1.089000],
            [1.026719, 0.000000, -0.363000],
            [-0.513360, -0.889165, -0.363000],
            [-0.513360, 0.889165, -0.363000],
        ]
        methane = Molecule(["C", "H", "H", "H", "H"], coords)
        element_types = get_element_list([methane])
        p2g = Molecule2Graph(element_types=element_types, cutoff=1.5)
        graph, state = p2g.get_graph(methane)
        # check the number of nodes
        assert np.allclose(graph.num_nodes(), 5)
        # check the number of edges
        assert np.allclose(graph.num_edges(), 8)
        # check the src_ids
        assert np.allclose(graph.edges()[0].numpy(), [0, 0, 0, 0, 1, 2, 3, 4])
        # check the dst_ids
        assert np.allclose(graph.edges()[1].numpy(), [1, 2, 3, 4, 0, 0, 0, 0])
        # check the atomic features of atom C
        assert np.allclose(graph.ndata["attr"][0], [0, 1])
        # check the atomic features of atom H
        assert np.allclose(graph.ndata["attr"][1], [1, 0])
        # check the shape of state features
        assert np.allclose(len(state), 2)
        # check the value of state features
        assert np.allclose(state, [3.208492, 0.8])
        # check the position of atom 0
        assert np.allclose(graph.ndata["pos"][0], [0.0, 0.0, 0.0])

    def test_get_graph_from_structure(self):
        structure_LiFePO4 = self.get_structure("LiFePO4")
        element_types = get_element_list([structure_LiFePO4])
        p2g = Structure2Graph(element_types=element_types, cutoff=4.0)
        graph, state = p2g.get_graph(structure_LiFePO4)
        # check the number of nodes
        assert np.allclose(graph.num_nodes(), structure_LiFePO4.num_sites)
        # check the atomic feature of atom 0
        assert np.allclose(graph.ndata["attr"][0].numpy(), [1, 0, 0, 0])
        # check the atomic feature of atom 4
        assert np.allclose(graph.ndata["attr"][4].numpy(), [0, 0, 0, 1])
        # check the number of bonds
        assert np.allclose(graph.num_edges(), 704)
        # check the state features
        assert np.allclose(state, [0.0, 0.0])
        structure_BaTiO3 = Structure.from_prototype("perovskite", ["Ba", "Ti", "O"], a=4.04)
        element_types = get_element_list([structure_BaTiO3])
        p2g = Structure2Graph(element_types=element_types, cutoff=4.0)
        graph, state = p2g.get_graph(structure_BaTiO3)
        # check the number of nodes
        assert np.allclose(graph.num_nodes(), structure_BaTiO3.num_sites)
        # check the atomic features of atom 0
        assert np.allclose(graph.ndata["attr"][0], [0, 0, 1])
        # check the atomic features of atom 1
        assert np.allclose(graph.ndata["attr"][1], [0, 1, 0])
        # check the number of edges
        assert np.allclose(graph.num_edges(), 76)
        # check the state features
        assert np.allclose(state, [0.0, 0.0])
        # check the position of atom 0
        assert np.allclose(graph.ndata["pos"][0], [0.0, 0.0, 0.0])
        # check the pbc offset from node 0 to image atom 6
        assert np.allclose(graph.edata["pbc_offset"][0], [-1, -1, -1])
        # check the lattice vector
        assert np.allclose(graph.edata["lattice"][0], [[4.04, 0.0, 0.0], [0.0, 4.04, 0.0], [0.0, 0.0, 4.04]])
        # check the volume
        assert np.allclose(graph.ndata["volume"][0], [65.939264])

    def test_get_graph_from_atoms(self):
        structure_LiFePO4 = self.get_structure("LiFePO4")
        element_types = get_element_list([structure_LiFePO4])
        adaptor = AseAtomsAdaptor()
        structure_ase = adaptor.get_atoms(structure_LiFePO4)
        p2g = Atoms2Graph(element_types=element_types, cutoff=4.0)
        graph, state = p2g.get_graph(structure_ase)
        # check the number of nodes
        assert np.allclose(graph.num_nodes(), len(structure_ase.get_atomic_numbers()))
        # check the atomic feature of atom 0
        assert np.allclose(graph.ndata["attr"][0].numpy(), [1, 0, 0, 0])
        # check the atomic feature of atom 4
        assert np.allclose(graph.ndata["attr"][4].numpy(), [0, 0, 0, 1])
        # check the number of bonds
        assert np.allclose(graph.num_edges(), 704)
        # check the state features
        assert np.allclose(state, [0.0, 0.0])

    def test_get_element_list(self):
        cscl = Structure.from_spacegroup("Pm-3m", Lattice.cubic(3), ["Cs", "Cl"], [[0, 0, 0], [0.5, 0.5, 0.5]])
        naf = Structure.from_spacegroup("Pm-3m", Lattice.cubic(3), ["Na", "F"], [[0, 0, 0], [0.5, 0.5, 0.5]])
        elem_list = get_element_list([cscl, naf])
        assert elem_list == ("F", "Na", "Cl", "Cs")


if __name__ == "__main__":
    unittest.main()



"""
Implementations of math functions.
"""

from __future__ import annotations

import os
from functools import lru_cache
from math import pi, sqrt

import numpy as np
import sympy
import torch
from scipy.optimize import brentq
from scipy.special import spherical_jn
from torch import nn

CWD = os.path.dirname(os.path.abspath(__file__))

"""
Precomputed Spherical Bessel function roots in a 2D array with dimension [128, 128]. The n-th (0-based index) root of
order l Spherical Bessel function is the `[l, n]` entry.
"""
SPHERICAL_BESSEL_ROOTS = torch.tensor(np.load(os.path.join(CWD, "sb_roots.npy")))


class GaussianExpansion(nn.Module):
    r"""
    Gaussian Radial Expansion.
    The bond distance is expanded to a vector of shape [m],
    where m is the number of Gaussian basis centers
    """

    def __init__(
        self,
        initial: float = 0.0,
        final: float = 4.0,
        num_centers: int = 20,
        width: None | float = 0.5,
    ):
        """
        Parameters
        ----------
        initial : float
                Location of initial Gaussian basis center.
        final : float
                Location of final Gaussian basis center
        number : int
                Number of Gaussian Basis functions
        width : float
                Width of Gaussian Basis functions
        """
        super().__init__()
        self.centers = nn.Parameter(torch.linspace(initial, final, num_centers), requires_grad=False)  # type: ignore
        if width is None:
            self.width = 1.0 / torch.diff(self.centers).mean()
        else:
            self.width = width

    def reset_parameters(self):
        """Reinitialize model parameters."""
        self.centers = nn.Parameter(self.centers, requires_grad=False)

    def forward(self, bond_dists):
        """Expand distances.

        Parameters
        ----------
        bond_dists :
            Bond (edge) distances between two atoms (nodes)

        Returns:
        -------
        A vector of expanded distance with shape [num_centers]
        """
        diff = bond_dists[:, None] - self.centers[None, :]
        return torch.exp(-self.width * (diff**2))


@lru_cache(maxsize=128)
def spherical_bessel_roots(max_l: int, max_n: int):
    """
    Calculate the spherical Bessel roots. The n-th root of the l-th
    spherical bessel function is the `[l, n]` entry of the return matrix.
    The calculation is based on the fact that the n-root for l-th
    spherical Bessel function `j_l`, i.e., `z_{j, n}` is in the range
    `[z_{j-1,n}, z_{j-1, n+1}]`. On the other hand we know precisely the
    roots for j0, i.e., sinc(x)

    Args:
        max_l: max order of spherical bessel function
        max_n: max number of roots
    Returns: root matrix of size [max_l, max_n]
    """
    temp_zeros = np.arange(1, max_l + max_n + 1) * pi  # j0
    roots = [temp_zeros[:max_n]]
    for i in range(1, max_l):
        roots_temp = []
        for j in range(max_n + max_l - i):
            low = temp_zeros[j]
            high = temp_zeros[j + 1]
            root = brentq(lambda x, v: spherical_jn(v, x), low, high, (i,))
            roots_temp.append(root)
        temp_zeros = np.array(roots_temp)
        roots.append(temp_zeros[:max_n])
    return np.array(roots)


class SphericalBesselFunction:
    """
    Calculate the spherical Bessel function based on sympy + pytorch implementations
    """

    def __init__(self, max_l: int, max_n: int = 5, cutoff: float = 5.0, smooth: bool = False):
        """
        Args:
            max_l: int, max order (excluding l)
            max_n: int, max number of roots used in each l
            cutoff: float, cutoff radius
            smooth: Whether to smooth the function.
        """
        self.max_l = max_l
        self.max_n = max_n
        self.cutoff = cutoff
        self.smooth = smooth
        if smooth:
            self.funcs = self._calculate_smooth_symbolic_funcs()
        else:
            self.funcs = self._calculate_symbolic_funcs()

    @lru_cache(maxsize=128)
    def _calculate_symbolic_funcs(self) -> list:
        """
        Spherical basis functions based on Rayleigh formula. This function
        generates
        symbolic formula.

        Returns: list of symbolic functions

        """
        x = sympy.symbols("x")
        funcs = [sympy.expand_func(sympy.functions.special.bessel.jn(i, x)) for i in range(self.max_l + 1)]
        return [sympy.lambdify(x, func, torch) for func in funcs]

    @lru_cache(maxsize=128)
    def _calculate_smooth_symbolic_funcs(self) -> list:
        return _get_lambda_func(max_n=self.max_n, cutoff=self.cutoff)

    def __call__(self, r):
        """
        Args:
            r: torch.tensor, distance tensor, 1D


        Returns: [n, max_n * max_l] spherical Bessel function results

        """
        if self.smooth:
            return self._call_smooth_sbf(r)
        return self._call_sbf(r)

    def _call_smooth_sbf(self, r):
        results = [i(r) for i in self.funcs]
        return torch.t(torch.stack(results))

    def _call_sbf(self, r):
        roots = SPHERICAL_BESSEL_ROOTS[: self.max_l, : self.max_n]

        results = []
        factor = torch.tensor(sqrt(2.0 / self.cutoff**3))
        for i in range(self.max_l):
            root = roots[i]
            func = self.funcs[i]
            func_add1 = self.funcs[i + 1]
            results.append(
                func(r[:, None] * root[None, :] / self.cutoff) * factor / torch.abs(func_add1(root[None, :]))
            )
        return torch.cat(results, axis=1)

    @staticmethod
    def rbf_j0(r, cutoff: float = 5.0, max_n: int = 3):
        """
        Spherical Bessel function of order 0, ensuring the function value
        vanishes at cutoff

        Args:
            r: torch.tensor pytorch tensors
            cutoff: float, the cutoff radius
            max_n: int max number of basis
        Returns: basis function expansion using first spherical Bessel function
        """
        n = (torch.arange(1, max_n + 1)).type(dtype=torch.float32)[None, :]
        r = r[:, None]
        return sqrt(2.0 / cutoff) * torch.sin(n * pi / cutoff * r) / r


def _y00(theta, phi):
    r"""
    Spherical Harmonics with `l=m=0`

    ..math::
        Y_0^0 = \frac{1}{2} \sqrt{\frac{1}{\pi}}

    Args:
        theta: torch.tensor, the azimuthal angle
        phi: torch.tensor, the polar angle

    Returns: `Y_0^0` results

    """
    return 0.5 * torch.ones_like(theta) * sqrt(1.0 / pi)


def _conjugate(x):
    return torch.conj(x)


class SphericalHarmonicsFunction:
    """
    Spherical Harmonics function
    """

    def __init__(self, max_l: int, use_phi: bool = True):
        """
        Args:
            max_l: int, max l (excluding l)
            use_phi: bool, whether to use the polar angle. If not,
                the function will compute `Y_l^0`
        """
        self.max_l = max_l
        self.use_phi = use_phi
        funcs = []
        theta, phi = sympy.symbols("theta phi")
        for lval in range(self.max_l):
            m_list = range(-lval, lval + 1) if self.use_phi else [0]  # type: ignore
            for m in m_list:
                func = sympy.functions.special.spherical_harmonics.Znm(lval, m, theta, phi).expand(func=True)
                funcs.append(func)
        # replace all theta with cos(theta)
        costheta = sympy.symbols("costheta")
        funcs = [i.subs({theta: sympy.acos(costheta)}) for i in funcs]
        self.orig_funcs = [sympy.simplify(i).evalf() for i in funcs]
        self.funcs = [sympy.lambdify([costheta, phi], i, [{"conjugate": _conjugate}, torch]) for i in self.orig_funcs]
        self.funcs[0] = _y00

    def __call__(self, costheta, phi=None):
        """
        Args:
            costheta: Cosine of the azimuthal angle
            phi: torch.tensor, the polar angle

        Returns: [n, m] spherical harmonic results, where n is the number
            of angles. The column is arranged following
            `[Y_0^0, Y_1^{-1}, Y_1^{0}, Y_1^1, Y_2^{-2}, ...]`
        """
        # costheta = torch.tensor(costheta, dtype=torch.complex64)
        # phi = torch.tensor(phi, dtype=torch.complex64)
        return torch.stack([func(costheta, phi) for func in self.funcs], axis=1)
        # results = results.type(dtype=DataType.torch_float)
        # return results


def _block_repeat(array, block_size, repeats):
    col_index = torch.arange(array.size()[1])
    indices = []
    start = 0

    for i, b in enumerate(block_size):
        indices.append(torch.tile(col_index[start : start + b], [repeats[i]]))
        start += b
    indices = torch.cat(indices, axis=0)
    return torch.index_select(array, 1, indices)


def combine_sbf_shf(sbf, shf, max_n: int, max_l: int, use_phi: bool):
    """
    Combine the spherical Bessel function and the spherical Harmonics function

    For the spherical Bessel function, the column is ordered by
        [n=[0, ..., max_n-1], n=[0, ..., max_n-1], ...], max_l blocks,

    For the spherical Harmonics function, the column is ordered by
        [m=[0], m=[-1, 0, 1], m=[-2, -1, 0, 1, 2], ...] max_l blocks, and each
        block has 2*l + 1
        if use_phi is False, then the columns become
        [m=[0], m=[0], ...] max_l columns

    Args:
        sbf: torch.tensor spherical bessel function results
        shf: torch.tensor spherical harmonics function results
        max_n: int, max number of n
        max_l: int, max number of l
        use_phi: whether to use phi
    Returns:
    """
    if sbf.size()[0] == 0:
        return sbf

    if not use_phi:
        repeats_sbf = torch.tensor([1] * max_l * max_n)
        block_size = [1] * max_l
    else:
        # [1, 1, 1, ..., 1, 3, 3, 3, ..., 3, ...]
        repeats_sbf = torch.tensor(np.repeat(2 * np.arange(max_l) + 1, repeats=max_n))
        # tf.repeat(2 * tf.range(max_l) + 1, repeats=max_n)
        block_size = 2 * np.arange(max_l) + 1  # type: ignore
        # 2 * tf.range(max_l) + 1
    expanded_sbf = torch.repeat_interleave(sbf, repeats_sbf, 1)
    expanded_shf = _block_repeat(shf, block_size=block_size, repeats=[max_n] * max_l)
    shape = max_n * max_l
    if use_phi:
        shape *= max_l
    return torch.reshape(expanded_sbf * expanded_shf, [-1, shape])


def _sinc(x):
    return torch.sin(x) / x


def spherical_bessel_smooth(r, cutoff: float = 5.0, max_n: int = 10):
    """
    This is an orthogonal basis with first
    and second derivative at the cutoff
    equals to zero. The function was derived from the order 0 spherical Bessel
    function, and was expanded by the different zero roots

    Ref:
        https://arxiv.org/pdf/1907.02374.pdf

    Args:
        r: torch.tensor distance tensor
        cutoff: float, cutoff radius
        max_n: int, max number of basis, expanded by the zero roots

    Returns: expanded spherical harmonics with derivatives smooth at boundary

    """
    n = torch.arange(max_n).type(dtype=torch.float32)[None, :]
    r = r[:, None]
    fnr = (
        (-1) ** n
        * sqrt(2.0)
        * pi
        / cutoff**1.5
        * (n + 1)
        * (n + 2)
        / torch.sqrt(2 * n**2 + 6 * n + 5)
        * (_sinc(r * (n + 1) * pi / cutoff) + _sinc(r * (n + 2) * pi / cutoff))
    )
    en = n**2 * (n + 2) ** 2 / (4 * (n + 1) ** 4 + 1)
    dn = [torch.tensor(1.0)]
    for i in range(1, max_n):
        dn.append(1 - en[0, i] / dn[-1])
    dn = torch.stack(dn)  # type: ignore
    gn = [fnr[:, 0]]
    for i in range(1, max_n):
        gn.append(1 / torch.sqrt(dn[i]) * (fnr[:, i] + torch.sqrt(en[0, i] / dn[i - 1]) * gn[-1]))

    return torch.t(torch.stack(gn))


@lru_cache(maxsize=128)
def _get_lambda_func(max_n, cutoff: float = 5.0):
    r = sympy.symbols("r")
    d0 = 1.0
    en = []
    for i in range(max_n):
        en.append(i**2 * (i + 2) ** 2 / (4 * (i + 1) ** 4 + 1))

    dn = [d0]
    for i in range(1, max_n):
        dn.append(1 - en[i] / dn[-1])

    fnr = []
    for i in range(max_n):
        fnr.append(
            (-1) ** i
            * sympy.sqrt(2.0)
            * sympy.pi
            / cutoff**1.5
            * (i + 1)
            * (i + 2)
            / sympy.sqrt(1.0 * (i + 1) ** 2 + (i + 2) ** 2)
            * (
                sympy.sin(r * (i + 1) * sympy.pi / cutoff) / (r * (i + 1) * sympy.pi / cutoff)
                + sympy.sin(r * (i + 2) * sympy.pi / cutoff) / (r * (i + 2) * sympy.pi / cutoff)
            )
        )

    gnr = [fnr[0]]
    for i in range(1, max_n):
        gnr.append(1 / sympy.sqrt(dn[i]) * (fnr[i] + sympy.sqrt(en[i] / dn[i - 1]) * gnr[-1]))
    return [sympy.lambdify([r], sympy.simplify(i), torch) for i in gnr]


def get_segment_indices_from_n(ns):
    """
    Get segment indices from number array. For example if
    ns = [2, 3], then the function will return [0, 0, 1, 1, 1]

    Args:
        ns: torch.tensor, the number of atoms/bonds array

    Returns:
        object:

    Returns: segment indices tensor
    """
    a = torch.arange(ns.size(dim=0))
    return a.repeat_interleave(ns, dim=0)


def get_range_indices_from_n(ns):
    """
    Give ns = [2, 3], return [0, 1, 0, 1, 2]

    Args:
        ns: torch.tensor, the number of atoms/bonds array

    Returns: range indices
    """
    max_n = torch.max(ns)
    n = ns.size(dim=0)
    n_range = torch.arange(max_n)
    matrix = n_range.tile(
        [n, 1],
    )
    mask = torch.arange(max_n)[None, :] < ns[:, None]

    #    return matrix[mask]
    return torch.masked_select(matrix, mask)


def repeat_with_n(ns, n):
    """
    Repeat the first dimension according to n array.

    Args:
        ns (torch.tensor): tensor
        n (torch.tensor): a list of replications

    Returns: repeated tensor

    """
    return torch.repeat_interleave(ns, n, dim=0)


def broadcast_states_to_bonds(g, state_feat):
    """
    Broadcast state attributes of shape [Ns, Nstate] to
    bond attributes shape [Nb, Nstate]

    Args:
        g: DGL graph
        state_feat: state_feature

    Returns: broadcasted state attributes
    """
    return state_feat.repeat((g.num_edges(), 1))


def broadcast_states_to_atoms(g, state_feat):
    """
    Broadcast state attributes of shape [Ns, Nstate] to
    bond attributes shape [Nb, Nstate]

    Args:
        g: DGL graph
        state_feat: state_feature

    Returns: broadcasted state attributes

    """
    return state_feat.repeat((g.num_nodes(), 1))


def scatter_sum(input_tensor: torch.tensor, segment_ids: torch.tensor, num_segments: int, dim: int) -> torch.tensor:
    """
    Scatter sum operation along the specified dimension. Modified from the
    torch_scatter library (https://github.com/rusty1s/pytorch_scatter).

    Args:
        input_tensor (torch.Tensor): The input tensor to be scattered.
        segment_ids (torch.Tensor): Segment ID for each element in the input tensor.
        num_segments (int): The number of segments.
        dim (int): The dimension along which the scatter sum operation is performed (default: -1).

    Returns:
        resulting tensor
    """
    segment_ids = broadcast(segment_ids, input_tensor, dim)
    size = list(input_tensor.size())
    if segment_ids.numel() == 0:
        size[dim] = 0
    else:
        size[dim] = num_segments
    output = torch.zeros(size, dtype=input_tensor.dtype)
    return output.scatter_add_(dim, segment_ids, input_tensor)


def unsorted_segment_fraction(data: torch.tensor, segment_ids: torch.tensor, num_segments: torch.tensor):
    """
    Segment fraction
    Args:
        data (torch.tensor): original data
        segment_ids (torch.tensor): segment ids
        num_segments (torch.tensor): number of segments
    Returns:
        data (torch.tensor): data after fraction
    """
    segment_sum = scatter_sum(input_tensor=data, segment_ids=segment_ids, dim=0, num_segments=num_segments)
    sums = torch.gather(segment_sum, 0, segment_ids)
    return torch.div(data, sums)


def broadcast(input_tensor: torch.tensor, target_tensor: torch.tensor, dim: int):
    """
    Broadcast input tensor along a given dimension to match the shape of the target tensor.
    Modified from torch_scatter library (https://github.com/rusty1s/pytorch_scatter).

    Args:
        input_tensor: The tensor to broadcast.
        target_tensor: The tensor whose shape to match.
        dim: The dimension along which to broadcast.

    Returns:
        resulting input tensor after broadcasting
    """
    if input_tensor.dim() == 1:
        for _ in range(0, dim):
            input_tensor = input_tensor.unsqueeze(0)
    for _ in range(input_tensor.dim(), target_tensor.dim()):
        input_tensor = input_tensor.unsqueeze(-1)
    target_shape = list(target_tensor.shape)
    target_shape[dim] = input_tensor.shape[dim]
    return input_tensor.expand(target_shape)



"""
Provides utilities for managing models and data.
"""
from __future__ import annotations

import inspect
import json
import logging
import os
from pathlib import Path

import requests
import torch

from matgl.config import MATGL_CACHE, PRETRAINED_MODELS_BASE_URL

logger = logging.getLogger(__file__)


class IOMixIn:
    """
    Mixin class for model saving and loading.
    """

    def save_args(self, locals: dict, kwargs: dict | None = None) -> None:
        r"""
        Method to save args into a private _init_args variable. This should be called after super in the __init__
        method, e.g., `self.save_args(locals(), kwargs)`.

        Args:
            locals: The result of locals().
            kwargs: kwargs passed to the class.
        """
        args = inspect.getfullargspec(self.__class__.__init__).args
        d = {k: v for k, v in locals.items() if k in args and k not in ("self", "__class__")}
        if kwargs is not None:
            d.update(kwargs)

        # If one of the args is a subclass of IOMixIn, we will serialize that class.
        for k, v in d.items():
            if issubclass(v.__class__, IOMixIn):
                d[k] = {"@class": v.__class__.__name__, "@module": v.__class__.__module__, "init_args": v._init_args}
        self._init_args = d

    def save(self, path: str | Path = ".", metadata: dict | None = None, makedirs: bool = True):
        """
        Save model to a directory. Three files will be saved.
        - path/model.pt, which contains the torch serialized model args.
        - path/state.pt, which contains the saved state_dict from the model.
        - path/model.json, a txt version of model.pt that is purely meant for ease of reference.

        Args:
            path: String or Path object to directory for model saving. Defaults to current working directory (".").
            metadata: Any additional metadata to be saved into the model.json file. For example, a good use would be
                a description of model purpose, the training set used, etc.
            makedirs: Whether to create the directory using os.makedirs(exist_ok=True). Note that if the directory
                already exists, makedirs will not do anything.
        """
        path = Path(path)
        if makedirs:
            os.makedirs(path, exist_ok=True)

        torch.save(self._init_args, path / "model.pt")  # type: ignore
        torch.save(self.state_dict(), path / "state.pt")  # type: ignore
        d = {
            "@class": self.__class__.__name__,
            "@module": self.__class__.__module__,
            "metadata": metadata,
            "kwargs": self._init_args,
        }  # type: ignore
        with open(path / "model.json", "w") as f:
            json.dump(d, f, default=lambda o: str(o), indent=4)

    @classmethod
    def load(cls, path: str | Path, include_json=False, **kwargs):
        """
        Load the model weights from a directory.

        Args:
            path (str|path): Path to saved model or name of pre-trained model. The search order is path, followed by
                download from PRETRAINED_MODELS_BASE_URL (with caching).
            include_json (bool): If True, the "model.json" file is also loaded. This file can contain metadata about
                the model, e.g., if scaling was performed in training the model, this file may contain the mean and
                standard deviation of the models, which needs to be applied to the final predictions.
            kwargs: Additional kwargs passed to RemoteFile class. E.g., a useful one might be force_download if you
                want to update the model.

        Returns: model_object if include_json is false. (model_object, dict) if include_json is True.
        """
        path = Path(path)

        fnames = ["model.pt", "state.pt"]
        if include_json:
            fnames.append("model.json")

        if all((path / fn).exists() for fn in fnames):
            fpaths = {fn: path / fn for fn in fnames}
        else:
            try:
                fpaths = {
                    fn: RemoteFile(f"{PRETRAINED_MODELS_BASE_URL}{path}/{fn}", **kwargs).local_path for fn in fnames
                }
            except BaseException:
                raise ValueError(
                    f"No valid model found in {path} or among pre-trained_models at "
                    f"{MATGL_CACHE} or {PRETRAINED_MODELS_BASE_URL}."
                ) from None

        if not torch.cuda.is_available():
            state = torch.load(fpaths["state.pt"], map_location=torch.device("cpu"))
        else:
            state = torch.load(fpaths["state.pt"])
        d = torch.load(fpaths["model.pt"])

        # Deserialize any args that are IOMixIn subclasses.
        for k, v in d.items():
            if isinstance(v, dict) and "@class" in v and "@module" in v:
                modname = v["@module"]
                classname = v["@class"]
                mod = __import__(modname, globals(), locals(), [classname], 0)
                cls_ = getattr(mod, classname)
                d[k] = cls_(**v["init_args"])
        d = {k: v for k, v in d.items() if not k.startswith("@")}
        model = cls(**d)
        model.load_state_dict(state)  # type: ignore

        if include_json:
            with open(fpaths["model.json"]) as f:
                model_data = json.load(f)

            return model, model_data

        return model


class RemoteFile:
    """
    Handling of download of remote files to a local cache.
    """

    def __init__(self, uri: str, cache_location: str | Path = MATGL_CACHE, force_download: bool = False):
        """

        Args:
            uri: Uniform resource identifier.
            cache_location: Directory to cache downloaded RemoteFile. By default, downloaded models are saved at
                $HOME/.matgl.
            force_download: To speed up access, a model with the same name in the cache location will be used if
                present. If you want to force a re-download, set this to True.
        """
        self.uri = uri
        toks = uri.split("/")
        self.model_name = toks[-2]
        self.fname = toks[-1]
        cache_location = Path(cache_location)
        os.makedirs(cache_location / self.model_name, exist_ok=True)
        self.local_path = cache_location / self.model_name / self.fname
        if (not self.local_path.exists()) or force_download:
            logger.info("Downloading from remote location...")
            self._download()
        else:
            logger.info(f"Using cached local file at {self.local_path}...")

    def _download(self):
        r = requests.get(self.uri, allow_redirects=True)
        with open(self.local_path, "wb") as f:
            f.write(r.content)

    def __enter__(self):
        """
        Support with context.

        Returns:
            Stream on local path.
        """
        self.stream = open(self.local_path, "rb")  # noqa: SIM115
        return self.stream

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exit the with context.

        Args:
            exc_type: Usual meaning in __exit__.
            exc_val: Usual meaning in __exit__.
            exc_tb: Usual meaning in __exit__.
        """
        self.stream.close()


def load_model(path: Path, **kwargs):
    r"""
    Convenience method to load a model from a directory or name.

    Args:
        path (str|path): Path to saved model or name of pre-trained model. The search order is path, followed by
            download from PRETRAINED_MODELS_BASE_URL (with caching).
        **kwargs: Additional kwargs passed to RemoteFile class. E.g., a useful one might be force_download if you
            want to update the model.

    Returns:
        Returns: model_object if include_json is false. (model_object, dict) if include_json is True.
    """
    path = Path(path)

    fnames = ["model.pt", "state.pt", "model.json"]

    if all((path / fn).exists() for fn in fnames):
        fpaths = {fn: path / fn for fn in fnames}
    else:
        try:
            fpaths = {fn: RemoteFile(f"{PRETRAINED_MODELS_BASE_URL}{path}/{fn}", **kwargs).local_path for fn in fnames}
        except BaseException:
            raise ValueError(
                f"No valid model found in {path} or among pre-trained_models at "
                f"{MATGL_CACHE} or {PRETRAINED_MODELS_BASE_URL}."
            ) from None

    with open(fpaths["model.json"]) as f:
        d = json.load(f)
        modname = d["@module"]
        classname = d["@class"]
        mod = __import__(modname, globals(), locals(), [classname], 0)
        cls_ = getattr(mod, classname)
        return cls_.load(path, **kwargs)



"""
Implementation of Interatomic Potentials
"""
from __future__ import annotations

import dgl
import numpy as np
import torch
from torch import nn
from torch.autograd import grad

from matgl.layers import AtomRef
from matgl.utils.io import IOMixIn


class Potential(nn.Module, IOMixIn):
    """
    A class representing an interatomic potential.
    """

    def __init__(
        self,
        model: nn.Module,
        data_mean: torch.tensor | None = None,
        data_std: torch.tensor | None = None,
        element_refs: np.ndarray | None = None,
        calc_forces: bool = True,
        calc_stresses: bool = True,
        calc_hessian: bool = False,
    ):
        """
        :param model: M3GNet model
        :param element_refs: Element reference values for each element
        :param calc_forces: Enable force calculations
        :param calc_stresses: Enable stress calculations
        :param calc_hessian: Enable hessian calculations
        """
        super().__init__()
        self.save_args(locals())
        self.model = model
        self.calc_forces = calc_forces
        self.calc_stresses = calc_stresses
        self.calc_hessian = calc_hessian
        self.element_refs: AtomRef | None
        if element_refs is not None:
            self.element_refs = AtomRef(property_offset=element_refs)
        else:
            self.element_refs = None

        self.data_mean = data_mean if data_mean is not None else torch.zeros(1)
        self.data_std = data_std if data_std is not None else torch.ones(1)

    def forward(
        self, g: dgl.DGLGraph, state_attr: torch.tensor | None = None, l_g: dgl.DGLGraph | None = None
    ) -> tuple:
        """
        Args:
            g: DGL graph
            state_attr: State attrs
            l_g: Line graph.

        Returns:
            energies, forces, stresses, hessian: torch.tensor
        """
        forces = torch.zeros(1)
        stresses = torch.zeros(1)
        hessian = torch.zeros(1)
        if self.calc_forces:
            g.ndata["pos"].requires_grad_(True)
        total_energies = self.data_std * self.model(g=g, state_attr=state_attr, l_g=l_g) + self.data_mean
        if self.element_refs is not None:
            property_offset = torch.squeeze(self.element_refs(g))
            total_energies += property_offset

        if self.calc_forces:
            grads = grad(
                total_energies,
                [g.ndata["pos"], g.edata["bond_vec"]],
                grad_outputs=torch.ones_like(total_energies),
                create_graph=True,
                retain_graph=True,
            )

            forces = -grads[0]
            if self.calc_hessian:
                r = -grads[0].view(-1)
                s = r.size(0)
                hessian = total_energies.new_zeros((s, s))
                for iatom in range(s):
                    tmp = grad([r[iatom]], g.ndata["pos"], retain_graph=iatom < s)[0]
                    if tmp is not None:
                        hessian[iatom] = tmp.view(-1)
        if self.calc_stresses:
            f_ij = -grads[1]
            stresses = []
            count_edge = 0
            count_node = 0
            for graph_id in range(g.batch_size):
                num_edges = g.batch_num_edges()[graph_id]
                num_nodes = 0
                stresses.append(
                    -1
                    * (
                        160.21766208
                        * torch.matmul(
                            g.edata["bond_vec"][count_edge : count_edge + num_edges].T,
                            f_ij[count_edge : count_edge + num_edges],
                        )
                        / g.ndata["volume"][count_node + num_nodes]
                    )
                )
                count_edge = count_edge + num_edges
                num_nodes = g.batch_num_nodes()[graph_id]
                count_node = count_node + num_nodes
            stresses = torch.cat(stresses)

        return total_energies, forces, stresses, hessian



"""
This package implements the layers for M*GNet.
"""
from __future__ import annotations

from matgl.layers._activations import SoftExponential, SoftPlus2
from matgl.layers._atom_ref import AtomRef
from matgl.layers._bond import BondExpansion
from matgl.layers._core import MLP, EdgeSet2Set, GatedMLP
from matgl.layers._embedding import EmbeddingBlock
from matgl.layers._graph_convolution import M3GNetBlock, M3GNetGraphConv, MEGNetBlock, MEGNetGraphConv
from matgl.layers._readout import ReduceReadOut, Set2SetReadOut, WeightedReadOut, WeightedReadOutPair
from matgl.layers._three_body import SphericalBesselWithHarmonics, ThreeBodyInteractions



"""
Cutoff functions for constructing M3GNet potentials.
"""

from __future__ import annotations

from math import pi

import torch


def polynomial_cutoff(r, cutoff: float):
    """
    Polynomial cutoff function
    Args:
        r (torch.tensor): radius distance tensor
        cutoff (float): cutoff distance

    Returns: polynomial cutoff functions

    """
    ratio = r / cutoff
    return torch.where(r <= cutoff, 1 - 6 * ratio**5 + 15 * ratio**4 - 10 * ratio**3, 0.0)


def cosine_cutoff(r: torch.Tensor, cutoff: float) -> torch.Tensor:
    """
    Cosine cutoff function
    Args:
        r (torch.tensor): radius distance tensor
        cutoff (float): cutoff distance

    Returns: cosine cutoff functions

    """
    return torch.where(r <= cutoff, 0.5 * (torch.cos(pi * r / cutoff) + 1), 0.0)



from __future__ import annotations

import unittest

import torch
from pymatgen.core.structure import Lattice, Structure
from torch import nn

from matgl.ext.pymatgen import Structure2Graph, get_element_list
from matgl.graph.compute import compute_pair_vector_and_distance
from matgl.layers import BondExpansion, EmbeddingBlock
from matgl.layers._readout import (
    ReduceReadOut,
    Set2SetReadOut,
    WeightedReadOut,
    WeightedReadOutPair,
)


class TestReadOut(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.s1 = Structure(Lattice.cubic(3.17), ["Mo", "S"], [[0, 0, 0], [0.5, 0.5, 0.5]])
        element_types = get_element_list([cls.s1])
        p2g = Structure2Graph(element_types=element_types, cutoff=4.0)
        graph, state = p2g.get_graph(cls.s1)
        cls.g1 = graph
        cls.state1 = state

        bond_vec, bond_dist = compute_pair_vector_and_distance(cls.g1)

        cls.g1.edata["bond_dist"] = bond_dist
        cls.g1.edata["bond_vec"] = bond_vec

    def test_weighted_readout(self):
        bond_vec, bond_dist = compute_pair_vector_and_distance(self.g1)
        bond_expansion = BondExpansion(rbf_type="SphericalBessel", max_n=3, max_l=3, cutoff=4.0, smooth=False)
        bond_basis = bond_expansion(self.g1.edata["bond_dist"])
        embed = EmbeddingBlock(
            degree_rbf=9, dim_node_embedding=16, dim_edge_embedding=16, dim_state_feats=16, activation=nn.SiLU()
        )
        node_attr = self.g1.ndata["attr"]
        edge_attr = bond_basis
        node_feat, edge_feat, state_feat = embed(node_attr, edge_attr, torch.tensor([1.0, 2.0]))
        self.g1.ndata["node_feat"] = node_feat
        self.g1.edata["edge_feat"] = edge_feat
        read_out = WeightedReadOut(in_feats=16, dims=[32, 32], num_targets=4)
        atomic_properties = read_out(self.g1)
        assert [atomic_properties.size(dim=0), atomic_properties.size(dim=1)] == [2, 4]

    def test_readout_pair(self):
        bond_expansion = BondExpansion(rbf_type="SphericalBessel", max_n=3, max_l=3, cutoff=4.0, smooth=False)
        bond_basis = bond_expansion(self.g1.edata["bond_dist"])
        node_attr = self.g1.ndata["attr"]
        edge_attr = bond_basis
        embed = EmbeddingBlock(
            degree_rbf=9, dim_node_embedding=16, dim_edge_embedding=16, dim_state_feats=16, activation=nn.SiLU()
        )
        node_feat, edge_feat, state_feat = embed(node_attr, edge_attr, torch.tensor([1.0, 2.0]))
        read_out = WeightedReadOutPair(in_feats=16, dims=[32, 32], num_targets=1)
        self.g1.ndata["node_feat"] = node_feat
        self.g1.edata["edge_feat"] = edge_feat
        pair_properties = read_out(self.g1)
        assert [pair_properties.size(dim=0), pair_properties.size(dim=1), pair_properties.size(dim=2)] == [2, 2, 1]

    def test_reduce_readout(self):
        bond_expansion = BondExpansion(rbf_type="SphericalBessel", max_n=3, max_l=3, cutoff=4.0, smooth=False)
        bond_basis = bond_expansion(self.g1.edata["bond_dist"])
        embed = EmbeddingBlock(
            degree_rbf=9, dim_node_embedding=16, dim_edge_embedding=16, dim_state_feats=16, activation=nn.SiLU()
        )
        node_attr = self.g1.ndata["attr"]
        edge_attr = bond_basis
        node_feat, edge_feat, state_feat = embed(node_attr, edge_attr, torch.tensor([1.0, 2.0]))
        self.g1.ndata["node_feat"] = node_feat
        self.g1.edata["edge_feat"] = edge_feat
        read_out = ReduceReadOut(op="mean", field="node_feat")
        output = read_out(self.g1)
        assert [output.size(dim=0), output.size(dim=1)] == [1, 16]
        read_out2 = ReduceReadOut(op="mean", field="edge_feat")
        output2 = read_out2(self.g1)
        assert [output2.size(dim=0), output2.size(dim=1)] == [1, 16]

    def test_set2set_readout(self):
        bond_expansion = BondExpansion(rbf_type="SphericalBessel", max_n=3, max_l=3, cutoff=4.0, smooth=False)
        bond_basis = bond_expansion(self.g1.edata["bond_dist"])
        embed = EmbeddingBlock(
            degree_rbf=9, dim_node_embedding=16, dim_edge_embedding=32, dim_state_feats=16, activation=nn.SiLU()
        )
        node_attr = self.g1.ndata["attr"]
        edge_attr = bond_basis
        node_feat, edge_feat, state_feat = embed(node_attr, edge_attr, torch.tensor([1.0, 2.0]))
        self.g1.ndata["node_feat"] = node_feat
        self.g1.edata["edge_feat"] = edge_feat
        read_out = Set2SetReadOut(
            num_steps=3,
            num_layers=3,
            field="node_feat",
        )
        output = read_out(self.g1)
        assert [output.size(dim=0), output.size(dim=1)] == [1, 32]
        read_out2 = Set2SetReadOut(
            num_steps=3,
            num_layers=3,
            field="edge_feat",
        )
        output2 = read_out2(self.g1)
        assert [output2.size(dim=0), output2.size(dim=1)] == [1, 64]


if __name__ == "__main__":
    unittest.main()



"""
Readout layer for M3GNet
"""

from __future__ import annotations

import dgl
import torch
from dgl.nn import Set2Set
from torch import nn

from matgl.layers._core import EdgeSet2Set, GatedMLP


class Set2SetReadOut(nn.Module):
    """
    The Set2Set readout function
    """

    def __init__(
        self,
        num_steps: int,
        num_layers: int,
        field: str,
    ):
        """
        Args:
            num_steps (int): Number of LSTM steps
            num_layers (int): Number of layers.
            field (str): Field of graph to perform the readout
        """
        super().__init__()
        self.field = field
        self.num_steps = num_steps
        self.num_layers = num_layers

    def forward(self, g: dgl.DGLGraph):
        s2s_kwargs = {"n_iters": self.num_steps, "n_layers": self.num_layers}
        if self.field == "node_feat":
            in_feats = g.ndata["node_feat"].size(dim=1)
            set2set = Set2Set(in_feats, **s2s_kwargs)
            out_tensor = set2set(g, g.ndata["node_feat"])
        elif self.field == "edge_feat":
            in_feats = g.edata["edge_feat"].size(dim=1)
            set2set = EdgeSet2Set(in_feats, **s2s_kwargs)
            out_tensor = set2set(g, g.edata["edge_feat"])
        return out_tensor


class ReduceReadOut(nn.Module):
    """
    Reduce atom or bond attributes into lower dimensional tensors as readout.
    This could be summing up the atoms or bonds, or taking the mean, etc.
    """

    def __init__(self, op: str = "mean", field: str = "node_feat"):
        """
        Args:
            op (str): op for the reduction
            field (str): Field of graph to perform the reduction
        """
        super().__init__()
        self.op = op
        self.field = field

    def forward(self, g: dgl.DGLGraph):
        """
        Args:
            g: DGL graph
        Returns:
            torch.tensor
        """
        if self.field == "node_feat":
            reduced_tensor = dgl.readout_nodes(g, feat="node_feat", op=self.op)
        elif self.field == "edge_feat":
            reduced_tensor = dgl.readout_edges(g, feat="edge_feat", op=self.op)
        return reduced_tensor


class WeightedReadOut(nn.Module):
    """
    Feed node features into Gated MLP as readout.
    """

    def __init__(self, in_feats: int, dims: list[int], num_targets: int):
        """
        Args:
           in_feats: input features (nodes)
           dims: NN architecture for Gated MLP
           num_targets: number of target properties
        """
        super().__init__()
        self.in_feats = in_feats
        self.dims = [in_feats, *dims, num_targets]
        self.gated = GatedMLP(in_feats=in_feats, dims=self.dims, activate_last=False)

    def forward(self, g: dgl.DGLGraph):
        """
        Args:
            g: DGL graph
        Returns:
            atomic_properties: torch.tensor
        """
        atomic_properties = self.gated(g.ndata["node_feat"])
        return atomic_properties


class WeightedReadOutPair(nn.Module):
    """
    Feed the average of atomic features i and j into weighted readout layer.
    """

    def __init__(self, in_feats, dims, num_targets, activation=None):
        super().__init__()
        self.in_feats = in_feats
        self.activation = activation
        self.num_targets = num_targets
        self.dims = [*dims, num_targets]
        self.gated = GatedMLP(in_feats=in_feats, dims=self.dims, activate_last=False)

    def forward(self, g: dgl.DGLGraph):
        num_nodes = g.ndata["node_feat"].size(dim=0)
        pair_properties = torch.zeros(num_nodes, num_nodes, self.num_targets)
        for i in range(num_nodes):
            for j in range(i, num_nodes):
                in_features = (g.ndata["node_feat"][i][:] + g.ndata["node_feat"][j][:]) / 2.0
                pair_properties[i][j][:] = self.gated(in_features)[:]
                pair_properties[j][i][:] = pair_properties[i][j][:]
        return pair_properties



from __future__ import annotations

import unittest

from pymatgen.core.structure import Lattice, Structure
from pymatgen.io.ase import AseAtomsAdaptor

from matgl.apps.pes import Potential
from matgl.ext.ase import M3GNetCalculator
from matgl.ext.pymatgen import get_element_list
from matgl.models import M3GNet


class TestAseInterface(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        s = Structure(Lattice.cubic(4.0), ["Mo", "S"], [[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]])
        cls.element_types = get_element_list([s])  # type: ignore
        adaptor = AseAtomsAdaptor()
        cls.s_ase = adaptor.get_atoms(s)  # type: ignore

    def test_M3GNetCalculator(self):
        model = M3GNet(element_types=self.element_types, is_intensive=False)
        ff = Potential(model=model)
        calc = M3GNetCalculator(potential=ff)
        self.s_ase.set_calculator(calc)
        assert [self.s_ase.get_potential_energy().size] == [1]
        assert list(self.s_ase.get_forces().shape) == [2, 3]
        assert list(self.s_ase.get_stress().shape) == [6]


if __name__ == "__main__":
    unittest.main()



"""
Implementations of multi-layer perceptron (MLP) and other helper classes.
"""
from __future__ import annotations

from typing import Callable

import torch
from dgl import DGLGraph, broadcast_edges, softmax_edges, sum_edges
from torch import nn
from torch.nn import LSTM, Linear, Module, ModuleList


class MLP(nn.Module):
    """
    An implementation of a multi-layer perceptron.
    """

    def __init__(
        self,
        dims: list[int],
        activation: Callable[[torch.Tensor], torch.Tensor] | None = None,
        activate_last: bool = False,
        bias_last: bool = True,
    ) -> None:
        """
        :param dims: Dimensions of each layer of MLP.
        :param activation: Activation function.
        :param activate_last: Whether to apply activation to last layer.
        :param bias_last: Whether to apply bias to last layer.
        """
        super().__init__()
        self._depth = len(dims) - 1
        self.layers = ModuleList()

        for i, (in_dim, out_dim) in enumerate(zip(dims[:-1], dims[1:])):
            if i < self._depth - 1:
                self.layers.append(Linear(in_dim, out_dim, bias=True))

                if activation is not None:
                    self.layers.append(activation)  # type: ignore
            else:
                self.layers.append(Linear(in_dim, out_dim, bias=bias_last))

                if activation is not None and activate_last:
                    self.layers.append(activation)  # type: ignore

    def __repr__(self):
        dims = []

        for layer in self.layers:
            if isinstance(layer, Linear):
                dims.append(f"{layer.in_features} \u2192 {layer.out_features}")
            else:
                dims.append(layer.__class__.__name__)

        return f'MLP({", ".join(dims)})'

    @property
    def last_linear(self) -> Linear | None:
        """
        :return: The last linear layer.
        """
        for layer in reversed(self.layers):
            if isinstance(layer, Linear):
                return layer
        return None

    @property
    def depth(self) -> int:
        """Returns depth of MLP."""
        return self._depth

    @property
    def in_features(self) -> int:
        """Return input features of MLP"""
        return self.layers[0].in_features

    @property
    def out_features(self) -> int:
        """Returns output features of MLP."""
        for layer in reversed(self.layers):
            if isinstance(layer, Linear):
                return layer.out_features
        raise RuntimeError

    def forward(self, inputs):
        """
        Applies all layers in turn.

        :param inputs: Input tensor
        :return: Output tensor
        """
        x = inputs
        for layer in self.layers:
            x = layer(x)

        return x


class GatedMLP(nn.Module):
    """
    An implementation of a Gated multi-layer perceptron.
    """

    def __init__(self, in_feats: int, dims: list[int], activate_last: bool = True, use_bias: bool = True):
        """
        :param in_feats: Dimension of input features.
        :param dims: Architecture of neural networks.
        :param activate_last: Whether applying activation to last layer or not.
        :param bias_last: Whether applying bias to last layer or not.
        """
        super().__init__()
        self.in_feats = in_feats
        self.dims = [in_feats, *dims]
        self._depth = len(dims)
        self.layers = nn.Sequential()
        self.gates = nn.Sequential()
        self.use_bias = use_bias
        self.activate_last = activate_last
        for i, (in_dim, out_dim) in enumerate(zip(self.dims[:-1], self.dims[1:])):
            if i < self._depth - 1:
                self.layers.append(nn.Linear(in_dim, out_dim, bias=use_bias))
                self.gates.append(nn.Linear(in_dim, out_dim, bias=use_bias))
                self.layers.append(nn.SiLU())
                self.gates.append(nn.SiLU())
            else:
                self.layers.append(nn.Linear(in_dim, out_dim, bias=use_bias))
                if self.activate_last:
                    self.layers.append(nn.SiLU())
                self.gates.append(nn.Linear(in_dim, out_dim, bias=use_bias))
                self.gates.append(nn.Sigmoid())

    def forward(self, inputs: torch.tensor):
        return self.layers(inputs) * self.gates(inputs)


class EdgeSet2Set(Module):
    """
    Implementation of Set2Set.
    """

    def __init__(self, input_dim: int, n_iters: int, n_layers: int) -> None:
        """
        :param input_dim: The size of each input sample.
        :param n_iters: The number of iterations.
        :param n_layers: The number of recurrent layers.
        """
        super().__init__()
        self.input_dim = input_dim
        self.output_dim = 2 * input_dim
        self.n_iters = n_iters
        self.n_layers = n_layers
        self.lstm = LSTM(self.output_dim, self.input_dim, n_layers)
        self.reset_parameters()

    def reset_parameters(self):
        """Reinitialize learnable parameters."""
        self.lstm.reset_parameters()

    def forward(self, g: DGLGraph, feat: torch.tensor):
        """
        Defines the computation performed at every call.

        :param g: Input graph
        :param feat: Input features.
        :return: One hot vector
        """
        with g.local_scope():
            batch_size = g.batch_size

            h = (
                feat.new_zeros((self.n_layers, batch_size, self.input_dim)),
                feat.new_zeros((self.n_layers, batch_size, self.input_dim)),
            )

            q_star = feat.new_zeros(batch_size, self.output_dim)

            for _ in range(self.n_iters):
                q, h = self.lstm(q_star.unsqueeze(0), h)
                q = q.view(batch_size, self.input_dim)
                e = (feat * broadcast_edges(g, q)).sum(dim=-1, keepdim=True)
                g.edata["e"] = e
                alpha = softmax_edges(g, "e")
                g.edata["r"] = feat * alpha
                readout = sum_edges(g, "r")
                q_star = torch.cat([q, readout], dim=-1)

            return q_star



from __future__ import annotations

import unittest

import numpy as np
import torch
from pymatgen.core.structure import Lattice, Molecule, Structure

from matgl.ext.pymatgen import Molecule2Graph, Structure2Graph, get_element_list
from matgl.graph.compute import (
    compute_pair_vector_and_distance,
    compute_theta_and_phi,
    create_line_graph,
)


def _loop_indices(bond_atom_indices, pair_dist, cutoff=4.0):
    bin_count = np.bincount(bond_atom_indices[:, 0], minlength=bond_atom_indices[-1, 0] + 1)
    indices = []
    start = 0
    for bcont in bin_count:
        for i in range(bcont):
            for j in range(bcont):
                if start + i == start + j:
                    continue
                if pair_dist[start + i] > cutoff or pair_dist[start + j] > cutoff:
                    continue
                indices.append([start + i, start + j])
        start += bcont
    return np.array(indices)


def _calculate_cos_loop(graph, threebody_cutoff=4.0):
    """
    Calculate the cosine theta of triplets using loops
    Args:
        graph: List
    Returns: a list of cosine theta values
    """

    _, _, n_sites = torch.unique(graph.edges()[0], return_inverse=True, return_counts=True)
    start_index = 0
    cos = []
    for n_site in n_sites:
        for i in range(n_site):
            for j in range(n_site):
                if i == j:
                    continue
                vi = graph.edata["bond_vec"][i + start_index].numpy()
                vj = graph.edata["bond_vec"][j + start_index].numpy()
                di = np.linalg.norm(vi)
                dj = np.linalg.norm(vj)
                if (di <= threebody_cutoff) and (dj <= threebody_cutoff):
                    cos.append(vi.dot(vj) / np.linalg.norm(vi) / np.linalg.norm(vj))
        start_index += n_site
    return cos


class TestCompute(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.s1 = Structure(Lattice.cubic(3.17), ["Mo", "Mo"], [[0.01, 0, 0], [0.5, 0.5, 0.5]])
        element_types = get_element_list([cls.s1])
        p2g = Structure2Graph(element_types=element_types, cutoff=5.0)  # type: ignore
        graph, state = p2g.get_graph(cls.s1)
        cls.g1 = graph
        cls.state1 = state

        coords = [
            [0.000000, 0.000000, 0.000000],
            [0.000000, 0.000000, 1.089000],
            [1.026719, 0.000000, -0.363000],
            [-0.513360, -0.889165, -0.363000],
            [-0.513360, 0.889165, -0.363000],
        ]
        methane = Molecule(["C", "H", "H", "H", "H"], coords)
        element_types = get_element_list([methane])
        p2g = Molecule2Graph(element_types=element_types, cutoff=2.0)  # type: ignore
        graph, state = p2g.get_graph(methane)
        cls.g2 = graph
        cls.state2 = state

    def test_compute_pair_vector(self):
        bv, bd = compute_pair_vector_and_distance(self.g1)
        self.g1.edata["bond_vec"] = bv
        d = torch.linalg.norm(self.g1.edata["bond_vec"], axis=1)

        _, _, _, d2 = self.s1.get_neighbor_list(r=5.0)

        np.testing.assert_array_almost_equal(np.sort(d), np.sort(d2))

    def test_compute_pair_vector_for_molecule(self):
        bv, bd = compute_pair_vector_and_distance(self.g2)
        self.g2.edata["bond_vec"] = bv
        d = torch.linalg.norm(self.g2.edata["bond_vec"], axis=1)

        d2 = np.array(
            [
                1.089,
                1.089,
                1.089,
                1.089,
                1.089,
                1.089,
                1.089,
                1.089,
                1.77833,
                1.77833,
                1.77833,
                1.77833,
                1.77833,
                1.77833,
                1.77833,
                1.77833,
                1.77833,
                1.77833,
                1.77833,
                1.77833,
            ]
        )

        np.testing.assert_array_almost_equal(np.sort(d), np.sort(d2))

    def test_compute_angle(self):
        bv, bd = compute_pair_vector_and_distance(self.g1)
        self.g1.edata["bond_vec"] = bv
        self.g1.edata["bond_dist"] = bd
        cos_loop = _calculate_cos_loop(self.g1, 4.0)

        line_graph = create_line_graph(self.g1, 4.0)
        line_graph.apply_edges(compute_theta_and_phi)
        np.testing.assert_array_almost_equal(
            np.sort(np.array(cos_loop)), np.sort(np.array(line_graph.edata["cos_theta"]))
        )

        bv, bd = compute_pair_vector_and_distance(self.g2)
        self.g2.edata["bond_vec"] = bv
        self.g2.edata["bond_dist"] = bd
        cos_loop = _calculate_cos_loop(self.g2, 2.0)

        line_graph = create_line_graph(self.g2, 2.0)
        line_graph.apply_edges(compute_theta_and_phi)
        np.testing.assert_array_almost_equal(
            np.sort(np.array(cos_loop)), np.sort(np.array(line_graph.edata["cos_theta"]))
        )


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import unittest

import numpy as np
import torch

from matgl.utils.cutoff import cosine_cutoff, polynomial_cutoff


class TestCutoff(unittest.TestCase):
    def test_polynomial(self):
        r = torch.linspace(1.0, 5.0, 11)
        three_body_cutoff = 4.0
        three_cutoff = polynomial_cutoff(r, three_body_cutoff)
        np.testing.assert_almost_equal(
            three_cutoff.tolist(),
            torch.tensor(
                [0.8965, 0.7648, 0.5931, 0.4069, 0.2352, 0.1035, 0.0266, 0.0012, 0.0000, 0.0000, 0.0000]
            ).tolist(),
            decimal=4,
        )

    def test_cosine(self):
        r = torch.linspace(1.0, 5.0, 11)
        three_body_cutoff = 4.0
        three_cutoff = cosine_cutoff(r, three_body_cutoff)
        np.testing.assert_almost_equal(
            three_cutoff.tolist(),
            torch.tensor([0.8536, 0.7270, 0.5782, 0.4218, 0.2730, 0.1464, 0.0545, 0.0062, 0.0000, 0.0000, 0.0000]),
            decimal=4,
        )


if __name__ == "__main__":
    unittest.main()
